#!/usr/bin/env python

import argparse
import sys
from math import log10
from os import mkdir, rename
from os.path import exists, join

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=(
        "Split sequences in a FASTA file into separate files, named "
        "either by their sequence id or numerically."
    ),
)

parser.add_argument("--outDir", default=".", help="The directory to make the files in.")

parser.add_argument(
    "--verbose",
    default=False,
    action="store_true",
    help="If given, print sequence ids as they are processed.",
)

parser.add_argument(
    "--numeric",
    default=False,
    action="store_true",
    help="If given, use numeric filenames, like 1.fasta, 2.fasta, etc.",
)

parser.add_argument(
    "--noLeadingZeroes",
    default=False,
    action="store_true",
    help="If given, numeric filenames will not have leading zeroes.",
)

parser.add_argument(
    "--force",
    default=False,
    action="store_true",
    help="If given, overwrite pre-existing files.",
)

parser.add_argument(
    "--saveAs",
    choices=("fasta", "fastq", "fasta-ss"),
    help=(
        "The output format. The default is to match the input format, "
        "so there is usually no need to specify this option. It can be "
        "used to force conversion from FASTQ to FASTA"
    ),
)

addFASTACommandLineOptions(parser)
args = parser.parse_args()
reads = parseFASTACommandLineOptions(args)

if not exists(args.outDir):
    mkdir(args.outDir)

saveAs = (
    args.saveAs
    or (args.fasta and "fasta")
    or (args.fastq and "fastq")
    or (args.fasta_ss and "fasta-ss")
)

# Note: we may be reading the FASTA input from stdin, so we cannot read it
# more than once (and I don't want to store it all because it may be very
# large). That's why we do a second phase of processing to renumber the
# files we created if --numeric is used (and --noLeadingZeroes is not).

count = 0
for count, read in enumerate(parseFASTACommandLineOptions(args), start=1):
    id_ = read.id.split()[0]
    if args.numeric:
        base = "%d.fasta" % count
    else:
        id_ = read.id.split()[0]
        base = "%s.fasta" % id_
    if args.verbose:
        print("Writing", base)
    filename = join(args.outDir, base)
    if not args.force and exists(filename):
        print(
            'Will not overwrite pre-existing file "%s". Use --force to '
            "make me. Exiting." % filename,
            file=sys.stderr,
        )
        sys.exit(1)
    with open(filename, "w") as fp:
        print(read.toString(saveAs), end="", file=fp)

# Rename numeric filenames to have leading zeroes.
if count and args.numeric and not args.noLeadingZeroes:
    width = int(log10(count)) + 1
    if width > 1:
        if args.verbose:
            print("Renaming to add leading zeroes")
        for i in range(1, count + 1):
            old = str(i)
            new = "%0*d" % (width, i)
            if old == new:
                # We've reached the point where the new names have as many
                # digits as the old, so we can stop.
                break
            else:
                oldFilename = "%s.fasta" % join(args.outDir, old)
                newFilename = "%s.fasta" % join(args.outDir, new)
                if not args.force and exists(newFilename):
                    print(
                        'Will not overwrite pre-existing file "%s". Use '
                        "--force to make me. Exiting." % newFilename,
                        file=sys.stderr,
                    )
                    sys.exit(1)
                if args.verbose:
                    print("  ", oldFilename, "->", newFilename)
                rename(oldFilename, newFilename)



#!/usr/bin/env python

import argparse
import sys

from dark.fasta import FastaReads
from dark.reads import Reads
from dark.sequence import findPrimerBidiLimits


def trimPrimers(primer: str, verbose: bool) -> None:
    """
    @param primer: A string sequence.
    @param verbose: A C{bool}, if C{True} output additional information about
        how often and where primers were found.
    """
    reads = Reads()
    absentCount = forwardCount = reverseCount = count = 0
    for seqRecord in FastaReads(sys.stdin):
        count += 1
        start, end = findPrimerBidiLimits(primer, seqRecord.sequence)
        if start == 0:
            if end == len(seqRecord):
                absentCount += 1
            else:
                reverseCount += 1
        else:
            forwardCount += 1
            if end != len(seqRecord):
                reverseCount += 1
        reads.add(seqRecord[start:end])

    if verbose:
        print(
            ("Read %d sequences. Found forward: %d, Found reversed: %d, Absent: %d")
            % (count, forwardCount, reverseCount, absentCount),
            file=sys.stderr,
        )

    reads.save(sys.stdout)


def main() -> None:
    parser = argparse.ArgumentParser(
        description=(
            "Given FASTA on stdin, look for a primer sequence "
            "and write trimmed FASTA (after the primer) to stdout."
        )
    )

    parser.add_argument("primer", help="The primer sequence.")
    parser.add_argument(
        "--verbose",
        type=bool,
        help="Print information on found primers.",
    )

    args = parser.parse_args()

    trimPrimers(args.primer.upper(), args.verbose)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import argparse
import re
import sys
from collections import defaultdict

import numpy as np
import polars as pl

from dark.filter import (
    addFASTAFilteringCommandLineOptions,
    parseFASTAFilteringCommandLineOptions,
)
from dark.reads import Reads
from dark.sam import SAMFilter, samfile


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Print SAM/BAM file coverage statistics.",
    )

    parser.add_argument(
        "--noFilter",
        action="store_true",
        dest="filter",
        default=False,
        help=(
            "Do not use our SAM filtering. This now the default since filtering "
            "can be extremely slow. Use --filter to turn filtering on. Note that "
            "if you use this option, any filtering option (other than --referenceId) "
            "you also specify that is provided by the SAMFilter.addFilteringOptions "
            "will be silently ignored!"
        ),
    )

    parser.add_argument(
        "--includeZeroes",
        action="store_true",
        help="Also write output for references that have no covering reads.",
    )

    parser.add_argument(
        "--filter",
        action="store_true",
        help="Use our SAM filtering. Note that this can be extremely slow.",
    )

    parser.add_argument(
        "--regex",
        action="store_true",
        help="Interpret the --referenceId arguments as regular expressions.",
    )

    parser.add_argument(
        "--match",
        action="store_true",
        help=(
            "Use 'match' when checking if references match regular expressions. "
            "This will cause the regular expression to be checked against the start "
            "of the reference name. If not given, the regular expression must "
            "match the entire reference name not just its prefix."
        ),
    )

    parser.add_argument(
        "--listReferences",
        action="store_true",
        help=(
            "Print the list of matching reference IDs and then exit. "
            "This makes it easy to adjust reference regular expressions "
            "to see what work would be done without doing it."
        ),
    )

    parser.add_argument(
        "--minReferenceLength",
        type=int,
        help=(
            "The minimum length reference to use. This can be useful if you give a "
            "reference regular expression that matches too many references and you "
            "want to also limit them by length. Or you can just give a minimal length "
            "and that way pick up all chromosomes without needing to name them."
        ),
    )

    parser.add_argument(
        "--tsv",
        help="The name of a TSV file to save results to.",
    )

    parser.add_argument(
        "--excel",
        help="The name of an Excel file to save results to.",
    )

    addFASTAFilteringCommandLineOptions(parser)
    SAMFilter.addFilteringOptions(parser, samfileIsPositional=True)

    return parser.parse_args()


def main() -> None:
    args = get_args()

    # We don't have a file of reads, we just want a read filter that we can use
    # to filter the SAM file query sequences and to get reference lengths from.
    reads = parseFASTAFilteringCommandLineOptions(args, Reads())
    samFilter = SAMFilter.parseFilteringOptions(args, reads.filterRead)

    if args.filter:
        filtering = True

        def filterRead(read):
            return not (read.is_del or read.is_refskip) and samFilter.filterAlignment(
                read.alignment
            )
    else:
        filtering = False

        def filterRead(read):
            return True

    referenceLengths = samFilter.referenceLengths(getAllReferences=True)

    if args.referenceId:
        if args.regex:
            regex = re.compile("|".join(map(re.escape, args.referenceId)))
            matcher = regex.match if args.match else regex.fullmatch
            wantedIds = set(id_ for id_ in referenceLengths if matcher(id_))

            if not wantedIds:
                sys.exit(
                    "None of the reference ids in the BAM file match the reference "
                    "id regular expression(s) you specified!"
                )

        else:
            wantedIds = set(args.referenceId)

            if not (wantedIds & set(referenceLengths)):
                sys.exit("None of the given reference ids occur in the input BAM file!")

    else:
        wantedIds = set(referenceLengths)

    if args.minReferenceLength is not None:
        n_before = len(wantedIds)
        wantedIds = set(
            id_ for id_ in wantedIds if referenceLengths[id_] >= args.minReferenceLength
        )

        if not wantedIds:
            sys.exit(
                f"None of the {n_before} potentially-included reference(s) has length "
                f"at least {args.minReferenceLength}."
            )

    if args.listReferences:
        print("\n".join(sorted(wantedIds)))
        sys.exit(0)

    coveredOffsetCount: dict[str, dict[int, int]] = defaultdict(
        lambda: defaultdict(int)
    )
    coveringReads = defaultdict(set)

    with samfile(args.samfile) as sam:
        for column in sam.pileup():
            referenceOffset = column.reference_pos
            for read in column.pileups:
                referenceId = read.alignment.reference_name
                if referenceId in wantedIds and (not filtering or filterRead(read)):
                    coveredOffsetCount[referenceId][referenceOffset] += 1
                    coveringReads[referenceId].add(read.alignment.query_name)

    rows = []

    for referenceId in sorted(wantedIds):
        offsetsCovered = len(coveredOffsetCount[referenceId])

        if offsetsCovered == 0 and not args.includeZeroes:
            continue

        referenceLength = referenceLengths[referenceId]

        nonZero = list(coveredOffsetCount[referenceId].values())
        assert len(nonZero) <= referenceLength
        coverageCounts = np.zeros(referenceLength, dtype=int)
        coverageCounts[: len(nonZero)] = nonZero

        rows.append(
            (
                referenceId,
                referenceLength,
                len(coveringReads[referenceId]),
                offsetsCovered,
                offsetsCovered / referenceLength,
                np.mean(coverageCounts),
                np.median(coverageCounts),
                np.min(coverageCounts),
                np.max(coverageCounts),
            )
        )

    if args.tsv or args.excel:
        schema = (
            "reference id",
            "reference length",
            "covering reads",
            "offsets covered",
            "fraction covered",
            "mean depth",
            "median depth",
            "min depth",
            "max depth",
        )

        df = pl.DataFrame(rows, schema=schema, orient="row")

        if args.tsv:
            df.write_csv(args.tsv, separator="\t")

        if args.excel:
            df.write_excel(args.excel)
    else:
        for row in rows:
            (
                referenceId,
                referenceLength,
                coveringReads,
                offsetsCovered,
                fraction,
                meanDepth,
                minDepth,
                maxDepth,
            ) = row

            print(
                f"{referenceId}: length {referenceLength}, covering "
                f"reads {coveringReads}, covered sites {offsetsCovered} "
                f"({fraction * 100.0:.4f%}%), mean coverage depth {meanDepth:.4f} "
                f"(min: {minDepth}, max: {maxDepth})"
            )


if __name__ == "__main__":
    main()



import sys
from collections import Counter, defaultdict
from contextlib import contextmanager
from json import dump, load
from subprocess import CalledProcessError
from typing import Iterator, Optional

import numpy as np
from pysam import (
    CDEL,
    CDIFF,
    CEQUAL,
    CHARD_CLIP,
    CINS,
    CMATCH,
    CPAD,
    CREF_SKIP,
    CSOFT_CLIP,
    AlignedRead,
    AlignmentFile,
)

from dark.fasta import FastaReads
from dark.process import Executor
from dark.reads import DNARead, Read
from dark.utils import pct


class SamError(Exception):
    "A SAM error."


class UnequalReferenceLengthError(SamError):
    "The references of interest in a SAM/BAM file are not of the same length."


class UnknownReference(SamError):
    "Reference sequence not found in SAM/BAM file."


class UnspecifiedReference(SamError):
    "Reference sequence not specified."


class ReferenceNameMismatchError(SamError):
    "Reference name mismatch."


class InvalidSAM(SamError):
    "SAM/BAM file has unexpected/invalid content."


# From https://samtools.github.io/hts-specs/SAMv1.pdf
CONSUMES_QUERY = {CMATCH, CINS, CSOFT_CLIP, CEQUAL, CDIFF}
CONSUMES_REFERENCE = {CMATCH, CDEL, CREF_SKIP, CEQUAL, CDIFF}


@contextmanager
def samfile(filename: str) -> Iterator[AlignmentFile]:
    """
    A context manager to open and close a SAM/BAM file.

    @param filename: A C{str} file name to open.
    """
    f = AlignmentFile(filename)
    yield f
    f.close()


def samtoolsInstalled() -> bool:
    """
    Test if samtools is installed.

    @return: A C{bool}, which is C{True} if DIAMOND seems to be installed.
    """
    try:
        Executor().execute("samtools help")
    except CalledProcessError:
        return False
    else:
        return True


def samReferences(filenameOrSamfile: str | AlignmentFile) -> list[str]:
    """
    List SAM/BAM file reference names.

    @param filenameOrSamfile: Either a C{str} SAM/BAM file name or an
        instance of C{pysam.AlignmentFile}.
    @return: A C{list} of C{str} reference names from the SAM file.
    """

    def _references(sam):
        return [sam.get_reference_name(i) for i in range(sam.nreferences)]

    if isinstance(filenameOrSamfile, str):
        with samfile(filenameOrSamfile) as sam:
            return _references(sam)
    else:
        return _references(filenameOrSamfile)


def samReferenceLengths(filenameOrSamfile: str | AlignmentFile) -> dict[str, int]:
    """
    Get SAM/BAM file reference ids and lengths.

    @param filenameOrSamfile: Either a C{str} SAM/BAM file name or an
        instance of C{pysam.AlignmentFile}.
    @return: A C{dict} mapping reference ids to lengths.
    """

    def _references(sam):
        result = {}
        for i in range(sam.nreferences):
            name = sam.get_reference_name(i)
            assert name not in result
            result[name] = sam.lengths[i]
        return result

    if isinstance(filenameOrSamfile, str):
        with samfile(filenameOrSamfile) as sam:
            return _references(sam)
    else:
        return _references(filenameOrSamfile)


def samReferencesToStr(filenameOrSamfile: str | AlignmentFile, indent: int = 0) -> str:
    """
    List SAM/BAM file reference names and lengths.

    @param filenameOrSamfile: Either a C{str} SAM/BAM file name or an
        instance of C{pysam.AlignmentFile}.
    @param indent: An C{int} number of spaces to indent each line.
    @return: A C{str} describing known reference names and their lengths.
    """
    indentStr = " " * indent

    def _references(sam):
        result = []
        for i in range(sam.nreferences):
            result.append(
                "%s%s (length %d)"
                % (indentStr, sam.get_reference_name(i), sam.lengths[i])
            )
        return "\n".join(result)

    if isinstance(filenameOrSamfile, str):
        with samfile(filenameOrSamfile) as sam:
            return _references(sam)
    else:
        return _references(filenameOrSamfile)


def _hardClip(sequence, quality, cigartuples):
    """
    Hard clip (if necessary) a sequence.

    @param sequence: A C{str} nucleotide sequence.
    @param quality: A C{str} quality string, or a C{list} of C{int} quality
        values as returned by pysam, or C{None} if the SAM file had a '*'
        for the quality string (which pysam converts to C{None}).
    @param cigartuples: An iterable of (operation, length) tuples, detailing
        the alignment, as per the SAM specification.
    @return: A 3-tuple consisting of
            1) a hard-clipped C{str} sequence if hard-clipping is indicated by
               the CIGAR operations.
            2) a hard-clipped quality C{str} or C{list} (depending on what
               type we were passed) if hard-clipping is indicated by the CIGAR
               operations.
            3) a Boolean, C{True} if hard clipping was performed by this
               function or C{False} if the hard clipping had already been
               done.
    """
    hardClipCount = cigarLength = 0
    for operation, length in cigartuples:
        hardClipCount += operation == CHARD_CLIP
        cigarLength += length if operation in CONSUMES_QUERY else 0

    sequenceLength = len(sequence)
    if quality is not None:
        assert sequenceLength == len(quality)
    clipLeft = clipRight = 0
    clippedSequence = sequence
    clippedQuality = quality

    if sequenceLength > cigarLength:
        alreadyClipped = False
    else:
        assert sequenceLength == cigarLength
        alreadyClipped = True

    if hardClipCount == 0:
        pass
    elif hardClipCount == 1:
        # Hard clip either at the start or the end.
        if cigartuples[0][0] == CHARD_CLIP:
            if not alreadyClipped:
                clipLeft = cigartuples[0][1]
                clippedSequence = sequence[clipLeft:]
                if quality is not None:
                    clippedQuality = quality[clipLeft:]
        elif cigartuples[-1][0] == CHARD_CLIP:
            if not alreadyClipped:
                clipRight = cigartuples[-1][1]
                clippedSequence = sequence[:-clipRight]
                if quality is not None:
                    clippedQuality = quality[:-clipRight]
        else:
            raise ValueError(
                "Invalid CIGAR tuples (%s) contains hard-clipping operation "
                "that is neither at the start nor the end of the sequence."
                % (cigartuples,)
            )
    elif hardClipCount == 2:
        # Hard clip at both the start and end.
        assert cigartuples[0][0] == cigartuples[-1][0] == CHARD_CLIP
        if not alreadyClipped:
            clipLeft, clipRight = cigartuples[0][1], cigartuples[-1][1]
            clippedSequence = sequence[clipLeft:-clipRight]
            if quality is not None:
                clippedQuality = quality[clipLeft:-clipRight]
    else:
        raise ValueError(
            "Invalid CIGAR tuples (%s) specifies hard-clipping %d times (2 "
            "is the maximum)." % (cigartuples, hardClipCount)
        )

    weClipped = bool(clipLeft or clipRight)

    if weClipped:
        assert not alreadyClipped
        if len(clippedSequence) + clipLeft + clipRight != sequenceLength:
            raise ValueError(
                "Sequence %r (length %d) clipped to %r (length %d), but the "
                "difference between these two lengths (%d) is not equal to "
                "the sum (%d) of the left and right clip lengths (%d and %d "
                "respectively). CIGAR tuples: %s"
                % (
                    sequence,
                    len(sequence),
                    clippedSequence,
                    len(clippedSequence),
                    abs(len(sequence) - len(clippedSequence)),
                    clipLeft + clipRight,
                    clipLeft,
                    clipRight,
                    cigartuples,
                )
            )
    else:
        assert len(clippedSequence) == sequenceLength
        if quality is not None:
            assert len(clippedQuality) == sequenceLength

    return clippedSequence, clippedQuality, weClipped


class SAMFilter:
    """
    Filter a SAM/BAM file.

    @param filename: The C{str} name of a BAM/SAM file to filter.
    @param filterRead: A function that takes a C{dark.reads.Read) instance
        and returns either C{None} or a C{Read} instance, according to
        whether the passed read should be omitted or not. If C{None} is
        passed, no filtering on reads (i.e., queries) is done.
    @param referenceIds: Either C{None} or a set of C{str} reference ids
        that should be kept (other references will be dropped).
    @param storeQueryIds: If C{True}, query ids will be stored (in
        C{self.queryIds}) as the SAM/BAM file is read.
    @param dropUnmapped: If C{True}, unmapped matches will be excluded.
    @param dropSecondary: If C{True}, secondary matches will be excluded.
    @param dropSupplementary: If C{True}, supplementary matches will be
        excluded.
    @param dropDuplicates: If C{True}, matches flagged as optical or PCR
        duplicates will be excluded.
    @param keepQCFailures: If C{True}, reads that are considered quality
        control failures will be included.
    @param minScore: If not C{None}, alignments with score tag values less
        than this value will not be output. If given, alignments that do not
        have a score will not be output.
    @param maxScore: If not C{None}, alignments with score tag values greater
        than this value will not be output. If given, alignments that do not
        have a score will not be output.
    @param scoreTag: The alignment tag to extract for minScore and maxScore
        comparisons.
    """

    def __init__(
        self,
        filename,
        filterRead=None,
        referenceIds=None,
        storeQueryIds=False,
        dropUnmapped=False,
        dropSecondary=False,
        dropSupplementary=False,
        dropDuplicates=False,
        keepQCFailures=False,
        minScore=None,
        maxScore=None,
        scoreTag="AS",
    ):
        self.filename = filename
        self.filterRead = filterRead
        self.referenceIds = referenceIds
        self.storeQueryIds = storeQueryIds
        self.dropUnmapped = dropUnmapped
        self.dropSecondary = dropSecondary
        self.dropSupplementary = dropSupplementary
        self.dropDuplicates = dropDuplicates
        self.keepQCFailures = keepQCFailures
        self.minScore = minScore
        self.maxScore = maxScore
        self.scoreTag = scoreTag

        # Detect when there are no filtering criteria, in which case
        # self.filterAlignment can return immediately.
        self.noFiltering = all(
            (
                filterRead is None,
                referenceIds is None,
                not any(
                    (
                        storeQueryIds,
                        dropUnmapped,
                        dropSecondary,
                        dropSupplementary,
                        dropDuplicates,
                        keepQCFailures,
                    )
                ),
                minScore is None,
                maxScore is None,
            )
        )

    @staticmethod
    def addFilteringOptions(
        parser,
        samfileIsPositional=False,
        samfileAction="store",
        samfileRequired=True,
        samfileNargs=None,
        referenceIdRequired=False,
    ):
        """
        Add options to an argument parser for filtering SAM/BAM.

        @param parser: An C{argparse.ArgumentParser} instance.
        @param samfileIsPositional: If C{True} the SAM/BAM file must
            be given as the final argument on the command line (without
            being preceded by --samfile).
        @param samfileAction: A C{str} action to take when 'samfile' arguments
            are found on the command line. Pass 'append' to allow multiple SAM
            files.
        @param samfileRequired: If C{True}, --samfile must be given on the
            command line. This is only relevant when samfileIsPositional is
            C{False} because positional arguments are always required. It may
            seem strange to allow no --samfile argument in a class that filters
            SAM files, but this option can be used in conjuction with scripts
            that can optionally take a SAM file (e.g.,
            genome-protein-summary.py).
        @param samfileNargs: The value to pass for 'nargs' in adding the
            samfile option.
        @param referenceIdRequired: If C{True}, make the --referenceId option
            required.
        """
        if samfileIsPositional:
            # Positional arguments are always required.
            assert samfileRequired, (
                "samfileIsPositional is True, so samfileRequired must also be True."
            )
            if samfileNargs is None:
                parser.add_argument(
                    "samfile", action=samfileAction, help="The SAM/BAM file to filter."
                )
            else:
                parser.add_argument(
                    "samfile",
                    action=samfileAction,
                    nargs=samfileNargs,
                    help="The SAM/BAM file to filter.",
                )
        else:
            if samfileNargs is None:
                parser.add_argument(
                    "--samfile",
                    required=samfileRequired,
                    action=samfileAction,
                    help="The SAM/BAM file to filter.",
                )
            else:
                parser.add_argument(
                    "--samfile",
                    required=samfileRequired,
                    nargs=samfileNargs,
                    action=samfileAction,
                    help="The SAM/BAM file to filter.",
                )

        parser.add_argument(
            "--referenceId",
            metavar="ID",
            action="append",
            required=referenceIdRequired,
            help=(
                "A reference sequence id whose alignments should be kept "
                "(alignments against other references will be dropped). "
                "If omitted, alignments against all references will be "
                "kept. May be repeated."
            ),
        )

        parser.add_argument(
            "--dropUnmapped",
            action="store_true",
            help="If given, unmapped matches will not be output.",
        )

        parser.add_argument(
            "--dropSecondary",
            action="store_true",
            help="If given, secondary matches will not be output.",
        )

        parser.add_argument(
            "--dropSupplementary",
            action="store_true",
            help="If given, supplementary matches will not be output.",
        )

        parser.add_argument(
            "--dropDuplicates",
            action="store_true",
            help=(
                "If given, matches flagged as optical or PCR duplicates "
                "will not be output."
            ),
        )

        parser.add_argument(
            "--keepQCFailures",
            action="store_true",
            help=(
                "If given, reads that are considered quality control "
                "failures will be included in the output."
            ),
        )

        parser.add_argument(
            "--minScore",
            type=float,
            metavar="FLOAT",
            help=(
                "If given, alignments with --scoreTag (default AS) values "
                "less than this value will not be output. If given, "
                "alignments that do not have a score will not be output."
            ),
        )

        parser.add_argument(
            "--maxScore",
            type=float,
            metavar="FLOAT",
            help=(
                "If given, alignments with --scoreTag (default AS) values "
                "greater than this value will not be output. If given, "
                "alignments that do not have a score will not be output."
            ),
        )

        parser.add_argument(
            "--scoreTag",
            default="AS",
            metavar="TAG",
            help=(
                "The alignment tag to extract for --minScore and --maxScore "
                "comparisons."
            ),
        )

    @classmethod
    def parseFilteringOptions(cls, args, filterRead=None, storeQueryIds=False):
        """
        Parse command line options (as added by C{self.addSAMFilteringOptions}).

        @param args: The command line arguments, as returned by
            C{argparse.parse_args}.
        @param filterRead: A one-argument function that accepts a read
            and returns C{None} if the read should be omitted in filtering
            or else a C{Read} instance.
        @param storeQueryIds: If C{True}, query ids will be stored as the
            SAM/BAM file is read.
        @return: A C{SAMFilter} instance.
        """
        return cls(
            args.samfile,
            filterRead=filterRead,
            referenceIds=set(args.referenceId) if args.referenceId else None,
            storeQueryIds=storeQueryIds,
            dropUnmapped=args.dropUnmapped,
            dropSecondary=args.dropSecondary,
            dropSupplementary=args.dropSupplementary,
            dropDuplicates=args.dropDuplicates,
            keepQCFailures=args.keepQCFailures,
            minScore=args.minScore,
            maxScore=args.maxScore,
        )

    def filterAlignment(self, alignment):
        """
        Test an alignment to see if it passes all filtering criteria.

        @param alignment: A pysam alignment instance.
        @return: A C{bool}, C{True} if the alignment passes our filtering,
            C{False} if it should be discarded.
        """
        if self.noFiltering:
            return True

        if self.minScore is not None or self.maxScore is not None:
            try:
                score = alignment.get_tag(self.scoreTag)
            except KeyError:
                return False
            else:
                if (self.minScore is not None and score < self.minScore) or (
                    self.maxScore is not None and score > self.maxScore
                ):
                    return False

        return (
            self.filterRead is None
            or self.filterRead(
                Read(alignment.query_name, alignment.query_sequence, alignment.qual)
            )
        ) and not (
            (self.referenceIds and alignment.reference_name not in self.referenceIds)
            or (alignment.is_unmapped and self.dropUnmapped)
            or (alignment.is_secondary and self.dropSecondary)
            or (alignment.is_supplementary and self.dropSupplementary)
            or (alignment.is_duplicate and self.dropDuplicates)
            or (alignment.is_qcfail and not self.keepQCFailures)
        )

    def alignments(self):
        """
        Get alignments from the SAM/BAM file, subject to filtering.

        @return: A generator that yields pysam alignment instances that pass
            our filtering criteria.
        """
        if self.storeQueryIds:
            queryIds = self.queryIds = set()
        else:
            queryIds = None

        lastAlignment = None
        count = 0
        with samfile(self.filename) as samAlignment:
            for count, alignment in enumerate(samAlignment.fetch(), start=1):
                if queryIds is not None:
                    queryIds.add(alignment.query_name)

                # Secondary and supplementary alignments may have a '*'
                # (pysam returns this as None) SEQ field, indicating that
                # the previous sequence should be used. This is best
                # practice according to section 2.5.2 of
                # https://samtools.github.io/hts-specs/SAMv1.pdf So we use
                # the last alignment query and quality strings if we get
                # None as a query sequence.
                if alignment.query_sequence is None:
                    if lastAlignment is None:
                        raise InvalidSAM(
                            "pysam produced an alignment (number %d) with no "
                            "query sequence without previously giving an "
                            "alignment with a sequence." % count
                        )
                    # Use the previous query sequence and quality. I'm not
                    # making the call to _hardClip dependent on
                    # alignment.cigartuples (as in the else clause below)
                    # because I don't think it's possible for
                    # alignment.cigartuples to be None in this case. If we
                    # have a second match on a query, then it must be
                    # aligned to something (i.e., it cannot be unmapped
                    # with no CIGAR string). The assertion will tell us if
                    # this is ever not the case.
                    assert alignment.cigartuples
                    (
                        alignment.query_sequence,
                        alignment.query_qualities,
                        _,
                    ) = _hardClip(
                        lastAlignment.query_sequence,
                        lastAlignment.query_qualities,
                        alignment.cigartuples,
                    )
                else:
                    lastAlignment = alignment
                    if alignment.cigartuples:
                        (
                            alignment.query_sequence,
                            alignment.query_qualities,
                            _,
                        ) = _hardClip(
                            alignment.query_sequence,
                            alignment.query_qualities,
                            alignment.cigartuples,
                        )

                if self.filterAlignment(alignment):
                    yield alignment

        self.alignmentCount = count

    def referenceLengths(self, getAllReferences: bool = False) -> dict[str, int]:
        """
        Get the lengths of wanted references.

        @param getAllReferences: Ignore self.referenceIds and get the lengths
            of all references.
        @raise UnknownReference: If a reference id is not present in the
            SAM/BAM file.
        @return: A C{dict} of C{str} reference id to C{int} length with a key
            for each reference id in C{self.referenceIds} or for all references
            if C{self.referenceIds} is C{None}.
        """
        result = {}

        with samfile(self.filename) as sam:
            if getAllReferences or not self.referenceIds:
                result = dict(zip(sam.references, sam.lengths))
            else:
                for referenceId in self.referenceIds:
                    tid = sam.get_tid(referenceId)
                    if tid == -1:
                        raise UnknownReference(
                            f"Reference {referenceId!r} is not present in the "
                            f"SAM/BAM file {self.filename!r}."
                        )
                    else:
                        result[referenceId] = sam.lengths[tid]

        return result


class PaddedSAM:
    """
    Obtain aligned (padded) queries from a SAM/BAM file.

    @param samFilter: A C{SAMFilter} instance.
    @raises UnequalReferenceLengthError: If C{referenceName} is C{None}
        and the reference sequence lengths in the SAM/BAM file are not all
        identical.
    @raises UnknownReference: If C{referenceName} does not exist.
    """

    def __init__(self, samFilter):
        referenceLengths = samFilter.referenceLengths()

        if len(set(referenceLengths.values())) != 1:
            raise UnequalReferenceLengthError(
                "Your %d SAM/BAM file reference sequence "
                "lengths (%s) are not all identical."
                % (
                    len(referenceLengths),
                    ", ".join(
                        "%s=%d" % (id_, referenceLengths[id_])
                        for id_ in sorted(referenceLengths)
                    ),
                )
            )

        # Get the length of any of the sequences (they are all identical).
        self.referenceLength = referenceLengths.popitem()[1]
        self.samFilter = samFilter
        # self.referenceInsertions will be keyed by query id (the query
        # that would cause a reference insertion). The values will be lists
        # of 2-tuples, with each 2-tuple containing an offset into the
        # reference sequence and the C{str} of nucleotides that would be
        # inserted starting at that offset.
        self.referenceInsertions = defaultdict(list)

    def queries(
        self,
        rcSuffix="",
        rcNeeded=False,
        padChar="-",
        queryInsertionChar="N",
        unknownQualityChar="!",
        allowDuplicateIds=False,
        addAlignment=False,
    ):
        """
        Produce padded (with gaps) queries according to the CIGAR string and
        reference sequence length for each matching query sequence.

        @param rcSuffix: A C{str} to add to the end of query names that are
            reverse complemented. This is added before the /1, /2, etc., that
            are added for duplicated ids (if there are duplicates and
            C{allowDuplicateIds} is C{False}.
        @param rcNeeded: If C{True}, queries that are flagged as matching when
            reverse complemented should have reverse complementing when
            preparing the output sequences. This must be used if the program
            that created the SAM/BAM input flags reversed matches but does not
            also store the reverse complemented query.
        @param padChar: A C{str} of length one to use to pad queries with to
            make them the same length as the reference sequence.
        @param queryInsertionChar:  A C{str} of length one to use to insert
            into queries when the CIGAR string indicates that the alignment
            of a query would cause a deletion in the reference. This character
            is inserted as a 'missing' query character (i.e., a base that can
            be assumed to have been lost due to an error) whose existence is
            necessary for the match to continue.
        @param unknownQualityChar: The character to put into the quality
            string when unknown bases are inserted in the query or the query
            is padded on the left/right with gaps.
        @param allowDuplicateIds: If C{True}, repeated query ids (due to
            secondary or supplemental matches) will not have /1, /2, etc.
            appended to their ids. So repeated ids may appear in the yielded
            FASTA.
        @param addAlignment: If C{True} the reads yielded by the returned
            generator will also have an C{alignment} attribute, being the
            C{pysam.AlignedSegment} for the query.
        @raises InvalidSAM: If a query has an empty SEQ field and either there
            is no previous alignment or the alignment is not marked as
            secondary or supplementary.
        @return: A generator that yields C{Read} instances that are padded
            with gap characters to align them to the length of the reference
            sequence. See C{addAlignment}, above, to yield reads with the
            corresponding C{pysam.AlignedSegment}.
        """
        referenceLength = self.referenceLength

        # Hold the count for each id so we can add /1, /2 etc to duplicate
        # ids (unless --allowDuplicateIds was given).
        idCount = Counter()

        MATCH_OPERATIONS = {CMATCH, CEQUAL, CDIFF}

        for lineNumber, alignment in enumerate(self.samFilter.alignments(), start=1):
            query = alignment.query_sequence
            quality = "".join(chr(q + 33) for q in alignment.query_qualities)

            if alignment.is_reverse:
                if rcNeeded:
                    query = DNARead("id", query).reverseComplement().sequence
                    quality = quality[::-1]
                if rcSuffix:
                    alignment.query_name += rcSuffix

            # Adjust the query id if it's a duplicate and we're not allowing
            # duplicates.
            if allowDuplicateIds:
                queryId = alignment.query_name
            else:
                count = idCount[alignment.query_name]
                idCount[alignment.query_name] += 1
                queryId = alignment.query_name + ("" if count == 0 else "/%d" % count)

            referenceStart = alignment.reference_start
            atStart = True
            queryIndex = 0
            referenceIndex = referenceStart
            alignedSequence = ""
            alignedQuality = ""

            for operation, length in alignment.cigartuples:
                # The operations are tested in the order they appear in
                # https://samtools.github.io/hts-specs/SAMv1.pdf It would be
                # more efficient to test them in order of frequency of
                # occurrence.
                if operation in MATCH_OPERATIONS:
                    atStart = False
                    alignedSequence += query[queryIndex : queryIndex + length]
                    alignedQuality += quality[queryIndex : queryIndex + length]
                elif operation == CINS:
                    # Insertion to the reference. This consumes query bases but
                    # we don't output them because the reference cannot be
                    # changed.  I.e., these bases in the query would need to be
                    # inserted into the reference.  Remove these bases from the
                    # query but record what would have been inserted into the
                    # reference.
                    atStart = False
                    self.referenceInsertions[queryId].append(
                        (
                            referenceIndex,
                            query[queryIndex : queryIndex + length],
                        )
                    )
                elif operation == CDEL:
                    # Delete from the reference. Some bases from the reference
                    # would need to be deleted to continue the match. So we put
                    # an insertion into the query to compensate.
                    atStart = False
                    alignedSequence += queryInsertionChar * length
                    alignedQuality += unknownQualityChar * length
                elif operation == CREF_SKIP:
                    # Skipped reference. Opens a gap in the query. For
                    # mRNA-to-genome alignment, an N operation represents an
                    # intron.  For other types of alignments, the
                    # interpretation of N is not defined. So this is unlikely
                    # to occur.
                    atStart = False
                    alignedSequence += queryInsertionChar * length
                    alignedQuality += unknownQualityChar * length
                elif operation == CSOFT_CLIP:
                    # Bases in the query that are not part of the match. We
                    # remove these from the query if they protrude before the
                    # start or after the end of the reference. According to the
                    # SAM docs, 'S' operations may only have 'H' operations
                    # between them and the ends of the CIGAR string.
                    if atStart:
                        # Don't set atStart=False, in case there's another 'S'
                        # operation.
                        unwantedLeft = length - referenceStart
                        if unwantedLeft > 0:
                            # The query protrudes left. Copy its right part.
                            alignedSequence += query[
                                queryIndex + unwantedLeft : queryIndex + length
                            ]
                            alignedQuality += quality[
                                queryIndex + unwantedLeft : queryIndex + length
                            ]
                            referenceStart = 0
                        else:
                            referenceStart -= length
                            alignedSequence += query[queryIndex : queryIndex + length]
                            alignedQuality += quality[queryIndex : queryIndex + length]
                    else:
                        unwantedRight = (
                            referenceStart + len(alignedSequence) + length
                        ) - referenceLength

                        if unwantedRight > 0:
                            # The query protrudes right. Copy its left part.
                            alignedSequence += query[
                                queryIndex : queryIndex + length - unwantedRight
                            ]
                            alignedQuality += quality[
                                queryIndex : queryIndex + length - unwantedRight
                            ]
                        else:
                            alignedSequence += query[queryIndex : queryIndex + length]
                            alignedQuality += quality[queryIndex : queryIndex + length]
                elif operation == CHARD_CLIP:
                    # Some bases have been completely removed from the query.
                    # This (H) can only be present as the first and/or last
                    # operation. There is nothing to do as the bases are simply
                    # not present in the query string in the SAM/BAM file.
                    pass
                elif operation == CPAD:
                    # This is "silent deletion from the padded reference",
                    # which consumes neither query nor reference.
                    atStart = False
                else:
                    raise ValueError("Unknown CIGAR operation:", operation)

                if operation in CONSUMES_QUERY:
                    queryIndex += length

                if operation in CONSUMES_REFERENCE:
                    referenceIndex += length

            if queryIndex != len(query):
                # Oops, we did not consume the entire query.
                raise ValueError(
                    "Query %r not fully consumed when parsing CIGAR string. "
                    "Query %r (len %d), final query index %d, CIGAR: %r"
                    % (
                        alignment.query_name,
                        query,
                        len(query),
                        queryIndex,
                        alignment.cigartuples,
                    )
                )

            # We cannot test we consumed the entire reference.  The CIGAR
            # string applies to (and exhausts) the query but is silent
            # about the part of the reference that lies to the right of the
            # aligned query.

            # Put gap characters before and after the aligned sequence so that
            # it is offset properly and matches the length of the reference.
            padRightLength = referenceLength - (referenceStart + len(alignedSequence))
            paddedSequence = (
                padChar * referenceStart + alignedSequence + padChar * padRightLength
            )
            paddedQuality = (
                unknownQualityChar * referenceStart
                + alignedQuality
                + unknownQualityChar * padRightLength
            )

            read = Read(queryId, paddedSequence, paddedQuality)

            if addAlignment:
                read.alignment = alignment

            yield read


class DistanceMatrix:
    """
    Maintain score (matching) information between reads and reference sequences
    as read from a SAM file (or files) and provide methods for calculating
    distance between references.
    """

    def __init__(self):
        # self.scores is a similarity matrix between references and
        # queries from a SAM file. The values are 1.0 or 0.0 when queries
        # match / do not match references unless we are using alignment
        # scores, as generated by tools like bowtie2 and bwa and stored in
        # a SAM tag. In that case, the alignment score value is stored (see
        # https://en.wikipedia.org/wiki/SAM_(file_format)#Optional_fields
        # for SAM tag info).
        #
        # The dictionary is keyed by str reference ids and the values are
        # dicts keyed by str query ids. The values of the sub-dictionaries
        # are the float scores (similarities). E.g.
        #
        # {
        #     'reference-1': {
        #         'query-1': 77.5,
        #         'query-2': 37.0,
        #     },
        #     'reference-2': {
        #         'query-1': 78.0,
        #     },
        # }
        #
        # If a reference/query id pair is not present in self.scores, it
        # means that query did not match that reference. The score method
        # (below) will assign such a pair a similarity score of 0.0.
        self.scores = {}

    def addFile(self, filename, scoreTag=None):
        """
        Add information from a SAM file.

        @param filename: A C{str} SAM file name to read.
        @param scoreTag: A C{str} alignment score tag to use instead of binary
            match / non-match values that are used if C{scoreTag} is C{None}.
            The score tag must be present in the optional SAM fields.
        """
        if scoreTag:

            def getScore(alignment, filename, count):
                """
                Get the alignment score from a SAM alignment.

                @param alignment: A samtools alignment.
                @param filename: The C{str} SAM file name being read.
                @param count: The C{int} line number in the SAM file.
                """
                try:
                    score = alignment.get_tag(scoreTag)
                except KeyError:
                    raise ValueError(
                        f"Alignment {count} in {filename!r} has no "
                        f"{scoreTag!r} score tag."
                    )
                else:
                    if score < 0.0:
                        raise ValueError(
                            f"Alignment {count} in {filename!r} has tag "
                            f"{scoreTag!r} with negative value ({score})."
                        )
                    return score

        else:

            def getScore(alignment, filename, count):
                """
                Return a binary score of 1.0 seeing as we know the query was
                mapped and no score tag is in use.

                @param args: ignored.
                """
                return 1.0

        scores = self.scores

        with samfile(filename) as samAlignment:
            for count, alignment in enumerate(samAlignment.fetch(), start=1):
                if alignment.is_unmapped:
                    continue

                queryId = alignment.query_name
                referenceId = alignment.reference_name
                score = getScore(alignment, filename, count)

                try:
                    preExisting = scores[referenceId][queryId]
                except KeyError:
                    if referenceId not in scores:
                        scores[referenceId] = {}

                    scores[referenceId][queryId] = score
                else:
                    if score > preExisting:
                        scores[referenceId][queryId] = score

    def score(self, referenceId, queryId):
        """
        Get the score for a reference and query.

        Although this method name does not start with an underscore, it's not
        likely to be used directly when this class is used to calculate
        distances between references according to query matching. The scores
        between references and queries (returned by this function) are just
        a component of the overall distance.

        @param referenceId: A C{str} reference id.
        @param queryId: A C{str} query id.
        @return: A C{float} score (zero if there was no match between the query
            and reference) or if the reference id or query id have never been
            seen. This could arguably cause a KeyError, but I decided against
            that to keep the code simpler / faster. If you pass in unknown
            query or reference ids, that's your problem (this should be very
            unlikely because our caller is operating on a SAM file, so the
            place they're most likely to be getting query and reference ids
            from to pass us is from that same SAM file.
        """
        try:
            return self.scores[referenceId][queryId]
        except KeyError:
            return 0.0

    def jaccardDistance(self, referenceId1, referenceId2):
        """
        Get the Jaccard distance between two references.  See
        https://en.wikipedia.org/wiki/Jaccard_index

        @param referenceId1: A C{str} reference id.
        @param referenceId2: A C{str} reference id.
        @return: A C{float} distance.
        """
        queryIds1 = set(self.scores.get(referenceId1, []))
        queryIds2 = set(self.scores.get(referenceId2, []))

        denominator = len(queryIds1 | queryIds2)

        if denominator:
            numerator = len(queryIds1 & queryIds2)
            similarity = numerator / denominator
            distance = 1.0 - similarity
            return distance
        else:
            # The references have no matching queries in common.
            return 1.0

    def soergelDistance(self, referenceId1, referenceId2):
        """
        Get the Soergel (i.e., weighted Jaccard) distance between two
        references. See https://en.wikipedia.org/wiki/Jaccard_index

        Note that this will be the same as the jaccardDistance unless a
        scoreTag was used in reading the SAM file.

        @param referenceId1: A C{str} reference id.
        @param referenceId2: A C{str} reference id.
        @return: A C{float} distance.
        """
        queryIds1 = set(self.scores.get(referenceId1, []))
        queryIds2 = set(self.scores.get(referenceId2, []))

        score = self.score
        scores = [
            (score(referenceId1, queryId), score(referenceId2, queryId))
            for queryId in (queryIds1 | queryIds2)
        ]

        denominator = sum(max(a, b) for a, b in scores)

        if denominator:
            numerator = sum(min(a, b) for a, b in scores)
            similarity = numerator / denominator
            distance = 1.0 - similarity
            return distance
        else:
            # Neither reference had a positive match against any query.
            return 1.0

    def matrix(
        self, referenceIds=None, metric="soergel", similarity=False, returnDict=False
    ):
        """
        Compute a distance matrix.

        @param referenceIds: An iterable of C{str} reference ids. If C{None},
            all known reference ids will be used.
        @param metric: A C{str}, either 'soergel' or 'jaccard'.
        @param similarity: If C{True}, return a similarity matrix.
        @param returnDict: If C{True}, return a C{dict} (see @return, below).
        @return: If C{returnDict} is C{False}, return a symmetric square
            C{np.array} of C{float} inter-reference distances (or similarities
            if C{similarity} is C{True}) in the range (0.0, 1.0). The order of
            the entries in the rows and columns of the returned array matches
            that of C{referenceIds} (if given) or the order in C{self.scores}
            otherwise.  If C{returnDict} is C{True}, return a C{dict} of
            C{dict}s, indexed by the two reference ids, with values as above.
        """
        assert metric in {"jaccard", "soergel"}
        referenceIds = tuple(referenceIds or self.scores)
        nIds = len(referenceIds)
        matrix = defaultdict(dict) if returnDict else np.empty((nIds, nIds))
        identityValue = 1.0 if similarity else 0.0
        func = self.jaccardDistance if metric == "jaccard" else self.soergelDistance

        if similarity:

            def metricFunc(ref1, ref2):
                return 1.0 - func(ref1, ref2)

        else:

            def metricFunc(ref1, ref2):
                return func(ref1, ref2)

        for index1, ref1 in enumerate(referenceIds):
            for index2, ref2 in enumerate(referenceIds):
                i, j = (ref1, ref2) if returnDict else (index1, index2)
                value = identityValue if i == j else metricFunc(ref1, ref2)
                matrix[i][j] = matrix[j][i] = value

        return matrix

    def save(self, fp):
        """
        Save the similarity matrix as JSON.

        @param fp: An open file pointer to write to.
        """
        dump(self.scores, fp, sort_keys=True, indent=4)

    def load(self, fp):
        """
        Load a similarity matrix from a JSON file.

        @param fp: An open file pointer to read from.
        """
        self.scores = load(fp)


def getReferenceInfo(
    bam, bamFilename, bamId=None, referenceFasta=None, fastaId=None, quiet=False
):
    """
    Get the id and length of a BAM file reference sequence.

    @param bam: An open BAM file.
    @param bamFilename: The C{str} file name of the BAM file.
    @param bamId: A C{str} BAM file reference name indicating which aligned
        reads to make a consensus from. If not given, will be inferred
        from the BAM file header.
    @param referenceFasta: A C{str} file name containing the sequence that was
        aligned to in making the BAM file.
    @param fastaId: A C{str} reference name indicating which sequence in
        C{referenceFasta} to use as a reference. Only considered if
        C{referenceFasta} is given. If not given and C{referenceFasta} is,
        the reference id will be inferred from reference names in the BAM
        header, or will be taken as the id of the first sequence in
        C{referenceFasta}.
    @param quiet: If C{True}, suppress diagnostic output.
    @raise UnknownReference: For an unknown reference.
    @raise UnspecifiedReference: If a reference is not given and cannot be
        inferred.
    @raise ReferenceNameMismatchError: If the names of the reference in the BAM
        file and the FASTA reference file (if given) do not match (unless both
        are explicitly given, in which case mismatching names are allowed as
        long as the sequence lengths are the same).
    @raise UnequalReferenceLengthError: If the lengths of the reference in the
        BAM file and the FASTA reference file (if given) do not match
    @return: A 2-tuple of C{str}, C{int} giving the refrence sequence id and
        length.
    """
    bamReferences = set(samReferences(bam))
    checkName = checkLength = True
    inferredBamId = bamId is None
    inferredFastaId = fastaId is None
    referenceLength = None

    if bamId is None:
        inferredBamId = True
    else:
        if bamId not in bamReferences:
            raise UnknownReference(
                f"BAM file {str(bamFilename)!r} does not mention a reference "
                f"with id {bamId!r}. Known references are: "
                f"{', '.join(sorted(bamReferences))}."
            )

        inferredBamId = False
        tid = bam.get_tid(bamId)
        referenceLength = bam.lengths[tid]

        if fastaId is not None:
            # Both ids have been given explicitly, so don't check that they
            # are identical, assume the user knows what they are doing (we
            # will just check the lengths, if necessary).
            checkName = False

    if referenceFasta is None:
        reference = None
    else:
        if fastaId is None:
            # We're given a reference FASTA file, but no id to use. Look at
            # the sequences in the FASTA and take the first one that has a
            # name and length matching a BAM reference (we could actually
            # look to see if there are more than one and, if so, give an
            # error if their sequences are not identical).
            firstRead = None
            for read in FastaReads(referenceFasta):
                if firstRead is None:
                    firstRead = read
                if read.id in bamReferences:
                    tid = bam.get_tid(read.id)
                    if len(read) == bam.lengths[tid]:
                        reference = read
                        checkLength = checkName = False
                        break
            else:
                # Try the first sequence in the FASTA file.
                reference = firstRead

            if firstRead is None:
                raise UnknownReference(
                    f"The FASTA reference file {str(referenceFasta)!r} "
                    f"contained no sequences."
                )
        else:
            for read in FastaReads(referenceFasta):
                if read.id == fastaId:
                    reference = read
                    break
            else:
                raise UnknownReference(
                    f"No sequence with id {fastaId!r} found in {str(referenceFasta)!r}."
                )

        # Set reference length if it has not already been set due to a
        # given bamId.
        if referenceLength is None:
            referenceLength = len(reference)

    if bamId is None:
        if len(bamReferences) == 1:
            # This is the only possibility.
            bamId = tuple(bamReferences)[0]
            tid = bam.get_tid(bamId)
            referenceLength = bam.lengths[tid]
        elif reference and reference.id in bamReferences:
            tid = bam.get_tid(reference.id)
            if len(reference) == bam.lengths[tid]:
                bamId = reference.id
                checkLength = checkName = False
        else:
            raise UnspecifiedReference(
                f"Could not infer a BAM reference. Available references are: "
                f"{', '.join(sorted(bamReferences))}."
            )

    if inferredBamId and not quiet:
        print(f"BAM reference id {bamId!r} inferred from context.", file=sys.stderr)

    if reference:
        if inferredFastaId and not quiet:
            print(
                f"FASTA reference id {reference.id!r} inferred from context.",
                file=sys.stderr,
            )

        if checkName and reference.id != bamId:
            raise ReferenceNameMismatchError(
                f"Reference FASTA sequence name {reference.id!r} does "
                f"not match the BAM id {bamId!r}."
            )

        if checkLength and len(reference) != referenceLength:
            raise UnequalReferenceLengthError(
                f"Reference FASTA sequence {reference.id!r} has length "
                f"{len(reference)}, but the BAM reference {bamId!r} has "
                f"length {referenceLength}."
            )

    return bamId, reference, referenceLength


class ReferenceReads:
    """
    Hold information about the reads that match a reference.

    @param id_: A C{str} reference id.
    @param length: The C{int} length of the reference.
    """

    def __init__(self, id_: str, length: int) -> None:
        assert length
        self.id_ = id_
        self.length = length
        self.coveredOffsets: set = set()
        self.duplicate: set = set()
        self.nonDuplicate: set = set()
        self.primary: set = set()
        self.qcFail: set = set()
        self.readIds: set = set()
        self.secondary: set = set()
        self.supplementary: set = set()

    def add(self, read: AlignedRead) -> None:
        """
        Add a matching read to this reference.

        @param read: An C{AlignedRead} to add.
        """
        id_ = read.query_name

        if read.is_secondary:
            self.secondary.add(id_)
        elif read.is_supplementary:
            self.supplementary.add(id_)
        else:
            self.primary.add(id_)

        # Note that it is possible that a SAM/BAM file has the same read id
        # more than once with identical flags (with the flags not including
        # 0x400 (1024, pysam.FDUP). Bowtie2 produces such lines. In such
        # cases, the read has not been flagged as a duplicate even though
        # it clearly is. To deal with this, if we see a read id again, we
        # unconditionally put it into self.duplicate (and make sure it is
        # not in self.nonDuplicate).
        if id_ in self.readIds:
            self.duplicate.add(id_)
            self.nonDuplicate.discard(id_)
        else:
            if read.is_duplicate:
                self.duplicate.add(id_)
            else:
                self.nonDuplicate.add(id_)

        self.readIds.add(id_)

        if read.is_qcfail:
            self.qcFail.add(id_)

        if read.reference_length is not None:
            self.coveredOffsets.update(
                range(
                    read.reference_start,
                    read.reference_start + read.reference_length,
                )
            )

    def coverage(self) -> float:
        """
        Determine the fraction of the reference covered by at least one read.

        @return: The C{float} coverage fraction.
        """
        return len(self.coveredOffsets) / self.length


class ReadsSummary:
    """
    Hold information about reads found in a BAM/SAM file.

    @param samFile: A C{str} filename to read.
    """

    def __init__(self, samFile: str) -> None:
        self.mapped: set[str] = set()
        self.unmapped: set[str] = set()
        self.readIds: set[str] = set()
        self.references: dict[str, ReferenceReads] = {}
        self.referenceLengths = SAMFilter(samFile).referenceLengths()
        self.sortedBy: Optional[str] = None
        self.sortedReferences: list[ReferenceReads] = []

        with samfile(samFile) as fp:
            for read in fp.fetch():
                id_ = read.query_name
                self.readIds.add(id_)

                if read.is_unmapped:
                    self.unmapped.add(id_)
                    continue

                self.mapped.add(id_)

                referenceName = read.reference_name
                try:
                    reference = self.references[referenceName]
                except KeyError:
                    reference = self.references[referenceName] = ReferenceReads(
                        referenceName, self.referenceLengths[referenceName]
                    )

                reference.add(read)

    def toString(
        self,
        excludeZeroes: bool = False,
        excludeIfNoAdditional: bool = False,
        sortBy: str = "count",
    ) -> str:
        """
        Return a string suitable for printing.

        @param excludeZeroes: Do not summarize references unmatched by any read.
        @param excludeIfNoAdditional: Do not summarize references that are mapped
            by no additional reads (i.e., reads other than those matching
            already-summarized references). This is useful C{sortBy} is 'count',
            making it possible to not summarize references only matched by reads
            matching references that have already been summarized
        @param sortBy: A C{str} indicating how to sort. Use 'count' to sort by
            decreasing read count (i.e., the reference with the highest number
            of matching reads comes first), or 'coverage' to sort by decreasing
            genome coverage, or 'name' to sort by reference name.
        @return: A C{list} of sorted C{ReferenceReads} instances.
        """
        totalReads = len(self.readIds)
        result = [
            "Found a total of %d read%s (%d mapped, %d unmapped) mapping "
            "against %d of %d reference%s."
            % (
                totalReads,
                "" if totalReads == 1 else "s",
                len(self.mapped),
                len(self.unmapped),
                len(self.references),
                len(self.referenceLengths),
                "" if len(self.referenceLengths) == 1 else "s",
            )
        ]

        if not self.references:
            result.append("No reads were mapped to any reference.")
            return "\n".join(result)

        cumulativeReadIds: set[str] = set()

        if self.sortedBy != sortBy:
            self.sortedReferences = self.sortReferences(sortBy)

        for count, reference in enumerate(self.sortedReferences, start=1):
            readIds = reference.readIds
            readCount = len(readIds)
            if readCount == 0 and excludeZeroes:
                continue
            newReadCount = len(readIds - cumulativeReadIds)
            if newReadCount == 0 and excludeIfNoAdditional:
                continue
            cumulativeReadIds.update(readIds)
            result.append(
                "\nReference %d: %s (%d nt):\n"
                "  Genome coverage: %.2f%%\n"
                "  Overall reads mapped to the reference: %s\n"
                "    Non-duplicates: %s, Duplicates: %s, QC fails: %s\n"
                "    Primary: %s, Secondary: %s, Supplementary: %s\n"
                "    Reads not matching any reference above: %s\n"
                "  Previously unmatched reads for this reference: %s"
                % (
                    count,
                    reference.id_,
                    reference.length,
                    reference.coverage() * 100.0,
                    pct(readCount, totalReads),
                    pct(len(reference.nonDuplicate), readCount),
                    pct(len(reference.duplicate), readCount),
                    pct(len(reference.qcFail), readCount),
                    pct(len(reference.primary), readCount),
                    pct(len(reference.secondary), readCount),
                    pct(len(reference.supplementary), readCount),
                    pct(newReadCount, totalReads),
                    pct(newReadCount, readCount),
                )
            )

        return "\n".join(result)

    def sortReferences(self, sortBy: str = "count"):
        """
        Sort the references.

        @param sortBy: A C{str} indicating how to sort. Use 'count' to sort by
            decreasing read count (i.e., the reference with the highest number
            of matching reads comes first), or 'coverage' to sort by decreasing
            genome coverage, or 'name' to sort by reference name.
        @raise AssertionError: If C{sortBy} has an unknown value.
        @return: A C{list} of sorted C{ReferenceReads} instances.
        """
        if self.sortedBy == sortBy:
            return self.sortedReferences

        if sortBy == "count":
            reverse = True

            def key(reference):
                return len(reference.readIds), reference.coverage()

        elif sortBy == "coverage":
            reverse = True

            def key(reference):
                return reference.coverage(), len(reference.readIds)

        elif sortBy == "name":
            reverse = False

            def key(reference):
                return reference.id_, len(reference.readIds), reference.coverage()

        else:
            raise ValueError(f"Unknown sortBy value {sortBy!r}.")

        self.sortedReferences = sorted(
            self.references.values(), key=key, reverse=reverse
        )
        self.sortedBy = sortBy

        return self.sortedReferences



#!/usr/bin/env python

import argparse
import sys
from os.path import basename, exists

from dark.process import Executor

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=(
        "Run bwa on a FASTA file. Optionally convert the result to "
        "BAM, sorting, and indexing."
    ),
)

parser.add_argument(
    "--bwaIndex",
    required=True,
    help="The name of the BWA index (created with bwa index).",
)

parser.add_argument(
    "--bwaArgs", default="mem", help="All arguments to be passed to BWA."
)

parser.add_argument(
    "--fastaFile",
    required=True,
    help=(
        "The name of the FASTA file whose sequences should be aligned with "
        "the BWA index."
    ),
)

parser.add_argument(
    "--base",
    help=(
        "The base name of files to create. Suffixes such as .sam and .bam "
        "will be added. If not given, the basename of the --fastaFile name "
        "will be used, stripped of its final suffix."
    ),
)

parser.add_argument(
    "--verbose",
    default=False,
    action="store_true",
    help="If specified, show the commands that were (or would be) executed.",
)

parser.add_argument(
    "--noBAM",
    default=False,
    action="store_true",
    help="If specified, do not convert SAM to BAM.",
)

parser.add_argument(
    "--noSort",
    default=False,
    action="store_true",
    help="If specified, do not sort the BAM.",
)

parser.add_argument(
    "--noIndex",
    default=False,
    action="store_true",
    help="If specified, do not index the sorted BAM.",
)

parser.add_argument(
    "--noClean",
    default=False,
    action="store_true",
    help="If specified, do not remove intermediate .sam and .bam files.",
)

parser.add_argument(
    "--force",
    default=False,
    action="store_true",
    help="If specified, overwrite pre-existing output files.",
)

parser.add_argument(
    "--dryRun",
    default=False,
    action="store_true",
    help="If specified, do not run commands, just print what would be done.",
)

args = parser.parse_args()

if args.base is None:
    fields = basename(args.fastaFile).rsplit(".", 1)
    if len(fields) < 2:
        print(
            "No --base argument was given and the --fastaFile argument "
            "does not have a .suffix that can be stripped.",
            file=sys.stderr,
        )
        sys.exit(1)
    base = fields[0]
else:
    base = args.base


samFile = base + ".sam"
bamFile = base + ".bam"
sortedBamFile = base + "-sorted.bam"

if not (args.force or args.dryRun):
    existing = []
    for filename in samFile, bamFile, sortedBamFile:
        if exists(filename):
            existing.append(filename)
    if existing:
        print(
            "Will not overwrite pre-existing file%s %s. "
            "Use --force to make me."
            % ("" if len(existing) == 1 else "s", ", ".join(existing)),
            file=sys.stderr,
        )
        sys.exit(2)

e = Executor(args.dryRun)

e.execute(
    "bwa %s '%s' '%s' > '%s'" % (args.bwaArgs, args.bwaIndex, args.fastaFile, samFile)
)

if not args.noBAM:
    e.execute("samtools view -b < '%s' > '%s'" % (samFile, bamFile))

    if not args.noClean:
        e.execute("rm '%s'" % samFile)

    if not args.noSort:
        e.execute("samtools sort '%s' > '%s'" % (bamFile, sortedBamFile))

        if not args.noClean:
            e.execute("rm '%s'" % bamFile)

        if not args.noIndex:
            e.execute("samtools index '%s'" % sortedBamFile)

if args.dryRun or args.verbose:
    print("\n".join(e.log))



import os
import re
from collections import Counter, defaultdict
from operator import itemgetter
from os.path import dirname, exists, join
from textwrap import fill
from urllib.parse import quote

import numpy as np

from dark.dimension import dimensionalIterator
from dark.fasta import FastaReads
from dark.fastq import FastqReads
from dark.filter import TitleFilter
from dark.html import NCBISequenceLinkURL
from dark.reads import Reads

# The following regex is deliberately greedy (using .*) to consume the
# whole protein name before backtracking to find the last [pathogen name]
# section. That way, it will match just the last [pathogen name] in a
# protein. This avoids situations in which two [...] delimited substrings
# are present in a protein name (in which case we just want the last).
# E.g., the following is a complete protein name:
#
#   gi|19919894|ref|NP_612577.1| Enzymatic polyprotein [Contains: Aspartic
#   protease; Endonuclease; Reverse transcriptase] [Carnation etched ring
#   virus]
#
# Unfortunately the regex doesn't find the pathogen name when the protein
# name has nested [...] sections, as in this example:
#
#   gi|224808893|ref|YP_002643049.1| replication-associated protein [Tomato
#   leaf curl Nigeria virus-[Nigeria:2006]]
#
# I decided not to worry about nested [...] sections (there are only 2
# instances that I know of).
_PATHOGEN_RE = re.compile(r"^(.*)\[([^\]]+)\]$")

# The pathogen name assigned to proteins whose id strings cannot be parsed
# for a pathogen name (see previous comment).  Do not use '<', '>' or any
# other HTML special chars in the following.
_NO_PATHOGEN_NAME = "[no pathogen name found in sequence id]"


def splitNames(names):
    """
    Split a sequence id string like "Protein name [pathogen name]" into two
    pieces using the final square brackets to delimit the pathogen name.

    @param names: A C{str} "protein name [pathogen name]" string.
    @return: A 2-C{tuple} giving the C{str} protein name and C{str} pathogen
        name. If C{names} cannot be split on square brackets, it is
        returned as the first tuple element, followed by _NO_PATHOGEN_NAME.
    """
    match = _PATHOGEN_RE.match(names)
    if match:
        proteinName = match.group(1).strip()
        pathogenName = match.group(2).strip()
    else:
        proteinName = names
        pathogenName = _NO_PATHOGEN_NAME

    return proteinName, pathogenName


def getPathogenProteinCounts(filenames):
    """
    Get the number of proteins for each pathogen in C{filenames}.

    @param filenames: Either C{None} or a C{list} of C{str} FASTA file names.
        If C{None} an empty C{Counter} is returned. If FASTA file names are
        given, their sequence ids should have the format used in the NCBI
        bacterial and viral protein reference sequence files, in which the
        protein name is followed by the pathogen name in square brackets.
    @return: A C{Counter} keyed by C{str} pathogen name, whose values are
        C{int}s with the count of the number of proteins for the pathogen.
    """
    result = Counter()
    if filenames:
        for filename in filenames:
            for protein in FastaReads(filename):
                _, pathogenName = splitNames(protein.id)
                if pathogenName != _NO_PATHOGEN_NAME:
                    result[pathogenName] += 1

    return result


class PathogenSampleFiles:
    """
    Maintain a cache of FASTA/FASTQ file names for the samples that contain a
    given pathogen, create de-duplicated (by read id) FASTA/FASTQ files
    for each pathogen/sample pair, provide functions to write out index files
    of sample and pathogen numbers (which are generated here in C{self.add}),
    and provide a filename lookup function for pathogen/sample combinations
    or just pathogen names by themselves.

    @param proteinGrouper: An instance of C{ProteinGrouper}.
    @param format_: A C{str}, either 'fasta' or 'fastq' indicating the format
        of the files containing the reads matching proteins.
    @raise ValueError: If C{format_} is unknown.
    """

    def __init__(self, proteinGrouper, format_="fasta"):
        self._proteinGrouper = proteinGrouper
        if format_ in ("fasta", "fastq"):
            self._format = format_
            self._readsClass = FastaReads if format_ == "fasta" else FastqReads
        else:
            raise ValueError("format_ must be either 'fasta' or 'fastq'.")
        self._pathogens = {}
        self._samples = {}
        self._readsFilenames = {}

    def add(self, pathogenName, sampleName):
        """
        Add a (pathogen name, sample name) combination and get its FASTA/FASTQ
        file name and unique read count. Write the FASTA/FASTQ file if it does
        not already exist. Save the unique read count into
        C{self._proteinGrouper}.

        @param pathogenName: A C{str} pathogen name.
        @param sampleName: A C{str} sample name.
        @return: A C{str} giving the FASTA/FASTQ file name holding all the
            reads (without duplicates, by id) from the sample that matched the
            proteins in the given pathogen.
        """
        pathogenIndex = self._pathogens.setdefault(pathogenName, len(self._pathogens))
        sampleIndex = self._samples.setdefault(sampleName, len(self._samples))

        try:
            return self._readsFilenames[(pathogenIndex, sampleIndex)]
        except KeyError:
            reads = Reads()
            for proteinMatch in self._proteinGrouper.pathogenNames[pathogenName][
                sampleName
            ]["proteins"].values():
                for read in self._readsClass(proteinMatch["readsFilename"]):
                    reads.add(read)
            saveFilename = join(
                proteinMatch["outDir"],
                "pathogen-%d-sample-%d.%s" % (pathogenIndex, sampleIndex, self._format),
            )
            reads.filter(removeDuplicatesById=True)
            nReads = reads.save(saveFilename, format_=self._format)
            # Save the unique read count into self._proteinGrouper
            self._proteinGrouper.pathogenNames[pathogenName][sampleName][
                "uniqueReadCount"
            ] = nReads
            self._readsFilenames[(pathogenIndex, sampleIndex)] = saveFilename
            return saveFilename

    def lookup(self, pathogenName, sampleName):
        """
        Look up a pathogen name, sample name combination and get its
        FASTA/FASTQ file name and unique read count.

        This method should be used instead of C{add} in situations where
        you want an exception to be raised if a pathogen/sample combination has
        not already been passed to C{add}.

        @param pathogenName: A C{str} pathogen name.
        @param sampleName: A C{str} sample name.
        @raise KeyError: If the pathogen name or sample name have not been
            seen, either individually or in combination.
        @return: A (C{str}, C{int}) tuple retrieved from self._readsFilenames
        """
        pathogenIndex = self._pathogens[pathogenName]
        sampleIndex = self._samples[sampleName]
        return self._readsFilenames[(pathogenIndex, sampleIndex)]

    def writeSampleIndex(self, fp):
        """
        Write a file of sample indices and names, sorted by index.

        @param fp: A file-like object, opened for writing.
        """
        print(
            "\n".join(
                "%d %s" % (index, name)
                for (index, name) in sorted(
                    (index, name) for (name, index) in self._samples.items()
                )
            ),
            file=fp,
        )

    def writePathogenIndex(self, fp):
        """
        Write a file of pathogen indices and names, sorted by index.

        @param fp: A file-like object, opened for writing.
        """
        print(
            "\n".join(
                "%d %s" % (index, name)
                for (index, name) in sorted(
                    (index, name) for (name, index) in self._pathogens.items()
                )
            ),
            file=fp,
        )

    def pathogenIndex(self, pathogenName):
        """
        Get the index for a pathogen.

        @param pathogenName: A C{str} pathogen name.
        @raise KeyError: If the named pathogen is unknown.
        @return: An C{int} giving the index for the pathogen (as set in
            self.add).
        """
        return self._pathogens[pathogenName]


class ProteinGrouper:
    """
    Group matched proteins by the pathogen they come from.

    @param assetDir: The C{str} directory name where
        C{noninteractive-alignment-panel.py} put its HTML, blue plot and
        alignment panel images, and FASTA or FASTQ files. This must be relative
        to the filenames that will later be passed to C{addFile}.
    @param sampleName: A C{str} sample name. This takes precedence over
        C{sampleNameRegex} (the two cannot be used together, obviously).
    @param sampleNameRegex: A C{str} regular expression that can be used to
        extract a short sample name from full file names subsequently passed
        to C{self.addFile}. The regular expression must have a matching group
        (delimited by parentheses) to capture the part of the file name that
        should be used as the sample name.
    @param format_: A C{str}, either 'fasta' or 'fastq' indicating the format
        of the files containing the reads matching proteins.
    @param proteinFastaFilenames: If not C{None}, a C{list} of C{str} filenames
        giving the name of the FASTA file with the protein AA sequences with
        their associated pathogens in square brackets. This is the format used
        by NCBI for the bacterial and viral reference sequence protein files.
        If given, the contents of this file will be used to determine how many
        proteins each matched pathogen has.
    @param saveReadLengths: If C{True}, save the lengths of all reads matching
        proteins.
    @param titleRegex: A regex that pathogen names must match.
        Note that this matching is done on the final part of the protein title
        in square brackets, according to the convention used by the NCBI viral
        refseq database and RVDB.
    @param negativeTitleRegex: A regex that pathogen names must not match.
        Note that this matching is done on the final part of the protein title
        in square brackets, according to the convention used by the NCBI viral
        refseq database and RVDB.
    @param pathogenDataDir: The C{str} directory where per-pathogen information
        (e.g., collected reads across all samples) should be written. Will be
        created (in C{self.toHTML}) if it doesn't exist.
    @raise ValueError: If C{format_} is unknown.
    """

    VIRALZONE = "https://viralzone.expasy.org/search?query="
    ICTV = "https://talk.ictvonline.org/search-124283882/?q="
    READCOUNT_MARKER = "*READ-COUNT*"
    READ_AND_HSP_COUNT_STR_SEP = "/"

    def __init__(
        self,
        assetDir="out",
        sampleName=None,
        sampleNameRegex=None,
        format_="fasta",
        proteinFastaFilenames=None,
        saveReadLengths=False,
        titleRegex=None,
        negativeTitleRegex=None,
        pathogenDataDir="pathogen-data",
    ):
        self._assetDir = assetDir
        self._sampleName = sampleName
        self._sampleNameRegex = re.compile(sampleNameRegex) if sampleNameRegex else None
        if format_ in ("fasta", "fastq"):
            self._format = format_
        else:
            raise ValueError("format_ must be either 'fasta' or 'fastq'.")
        self._saveReadLengths = saveReadLengths

        if titleRegex or negativeTitleRegex:
            self.titleFilter = TitleFilter(
                positiveRegex=titleRegex, negativeRegex=negativeTitleRegex
            )
        else:
            self.titleFilter = None

        self._pathogenDataDir = pathogenDataDir

        self._pathogenProteinCount = getPathogenProteinCounts(proteinFastaFilenames)

        # pathogenNames will be a dict of dicts of dicts. The first two keys
        # will be a pathogen name and a sample name. The final dict will
        # contain 'proteins' (a list of dicts) and 'uniqueReadCount' (an int).
        self.pathogenNames = {}
        # sampleNames is keyed by sample name and will have values that hold
        # the sample's alignment panel index.html file.
        self.sampleNames = {}
        self.pathogenSampleFiles = PathogenSampleFiles(self, format_=format_)

    def _title(self):
        """
        Create a title summarizing the pathogens and samples.

        @return: A C{str} title.
        """
        return "Overall, proteins from %d pathogen%s were found in %d sample%s." % (
            len(self.pathogenNames),
            "" if len(self.pathogenNames) == 1 else "s",
            len(self.sampleNames),
            "" if len(self.sampleNames) == 1 else "s",
        )

    def addFile(self, filename, fp):
        """
        Read and record protein information for a sample.

        @param filename: A C{str} file name.
        @param fp: An open file pointer to read the file's data from.
        @raise ValueError: If information for a pathogen/protein/sample
            combination is given more than once.
        """
        if self._sampleName:
            sampleName = self._sampleName
        elif self._sampleNameRegex:
            match = self._sampleNameRegex.search(filename)
            if match:
                sampleName = match.group(1)
            else:
                sampleName = filename
        else:
            sampleName = filename

        outDir = join(dirname(filename), self._assetDir)

        self.sampleNames[sampleName] = join(outDir, "index.html")

        for index, proteinLine in enumerate(fp):
            proteinLine = proteinLine[:-1]
            (
                coverage,
                medianScore,
                bestScore,
                readCount,
                hspCount,
                proteinLength,
                names,
            ) = proteinLine.split(None, 6)

            proteinName, pathogenName = splitNames(names)

            # Ignore pathogens with names we don't want.
            if (
                self.titleFilter
                and self.titleFilter.accept(pathogenName) == TitleFilter.REJECT
            ):
                continue

            if pathogenName not in self.pathogenNames:
                self.pathogenNames[pathogenName] = {}

            if sampleName not in self.pathogenNames[pathogenName]:
                self.pathogenNames[pathogenName][sampleName] = {
                    "proteins": {},
                    "uniqueReadCount": None,
                }

            proteins = self.pathogenNames[pathogenName][sampleName]["proteins"]

            # We should only receive one line of information for a given
            # pathogen/sample/protein combination.
            if proteinName in proteins:
                raise ValueError(
                    "Protein %r already seen for pathogen %r sample %r."
                    % (proteinName, pathogenName, sampleName)
                )

            readsFilename = join(outDir, "%d.%s" % (index, self._format))

            if proteinName.count("|") < 5:
                # Assume this is an NCBI refseq id, like
                # YP_009137153.1 uracil glycosylase [Human alphaherpesvirus 2]
                # with a protein but not a genome accession.
                proteinURL = NCBISequenceLinkURL(proteinName, field=0, delim=" ")
                genomeURL = None
            else:
                # Assume this is an RVDB id, like
                # acc|GENBANK|ABJ91970.1|GENBANK|DQ876317|pol protein [HIV]
                # with both protein and genome accession numbers.
                proteinURL = NCBISequenceLinkURL(proteinName, field=2)
                genomeURL = NCBISequenceLinkURL(proteinName, field=4)

            proteinInfo = proteins[proteinName] = {
                "bestScore": float(bestScore),
                "bluePlotFilename": join(outDir, "%d.png" % index),
                "coverage": float(coverage),
                "readsFilename": readsFilename,
                "hspCount": int(hspCount),
                "index": index,
                "medianScore": float(medianScore),
                "outDir": outDir,
                "proteinLength": int(proteinLength),
                "proteinName": proteinName,
                "proteinURL": proteinURL,
                "genomeURL": genomeURL,
                "readCount": int(readCount),
            }

            if proteinInfo["readCount"] == proteinInfo["hspCount"]:
                proteinInfo["readAndHspCountStr"] = readCount
            else:
                proteinInfo["readAndHspCountStr"] = "%s%s%s" % (
                    readCount,
                    self.READ_AND_HSP_COUNT_STR_SEP,
                    hspCount,
                )

            if self._saveReadLengths:
                readsClass = FastaReads if self._format == "fasta" else FastqReads
                proteins[proteinName]["readLengths"] = tuple(
                    len(read) for read in readsClass(readsFilename)
                )

    def _computeUniqueReadCounts(self):
        """
        Add all pathogen / sample combinations to self.pathogenSampleFiles.

        This will make all de-duplicated (by id) FASTA/FASTQ files and store
        the number of de-duplicated reads into C{self.pathogenNames}.
        """
        for pathogenName, samples in self.pathogenNames.items():
            for sampleName in samples:
                self.pathogenSampleFiles.add(pathogenName, sampleName)

    def toStr(self, title="Summary of pathogens", preamble=None):
        """
        Produce a string representation of the pathogen summary.

        @param title: The C{str} title for the output.
        @param preamble: The C{str} descriptive preamble, or C{None} if no
            preamble is needed.
        @return: A C{str} suitable for printing.
        """
        # Note that the string representation contains much less
        # information than the HTML summary. E.g., it does not contain the
        # unique (de-duplicated, by id) read count, since that is only computed
        # when we are making combined FASTA files of reads matching a
        # pathogen.
        readCountGetter = itemgetter("readCount")
        result = []
        append = result.append

        result.extend((title, ""))
        if preamble:
            result.extend((preamble, ""))
        result.extend((self._title(), ""))

        for pathogenName in sorted(self.pathogenNames):
            samples = self.pathogenNames[pathogenName]
            sampleCount = len(samples)
            append(
                "%s (in %d sample%s)"
                % (pathogenName, sampleCount, "" if sampleCount == 1 else "s")
            )
            for sampleName in sorted(samples):
                proteins = samples[sampleName]["proteins"]
                proteinCount = len(proteins)
                totalReads = sum(readCountGetter(p) for p in proteins.values())
                append(
                    "  %s (%d protein%s, %d read%s)"
                    % (
                        sampleName,
                        proteinCount,
                        "" if proteinCount == 1 else "s",
                        totalReads,
                        "" if totalReads == 1 else "s",
                    )
                )
                for proteinName in sorted(proteins):
                    append(
                        "    %(coverage).2f\t%(medianScore).2f\t"
                        "%(bestScore).2f\t%(readAndHspCountStr)11s\t"
                        "%(proteinName)s" % proteins[proteinName]
                    )
            append("")

        return "\n".join(result)

    def toHTML(
        self,
        pathogenPanelFilename=None,
        minProteinFraction=0.0,
        minProteinCount=0,
        pathogenType="viral",
        title="Summary of pathogens",
        preamble=None,
        sampleIndexFilename=None,
        pathogenIndexFilename=None,
        omitVirusLinks=False,
        omitSampleProteinCount=False,
    ):
        """
        Produce an HTML string representation of the pathogen summary.

        @param pathogenPanelFilename: If not C{None}, a C{str} filename to
            write a pathogen panel PNG image to.
        @param minProteinFraction: The C{float} minimum fraction of proteins
            in a pathogen that must be matched by a sample in order for that
            pathogen to be displayed for that sample.
        @param minProteinCount: The C{int} minimum number of proteins
            in a pathogen that must be matched by a sample in order for that
            pathogen to be displayed for that sample.
        @param pathogenType: A C{str} giving the type of the pathogen involved,
            either 'bacterial' or 'viral'.
        @param title: The C{str} title for the HTML page.
        @param preamble: The C{str} descriptive preamble for the HTML page, or
            C{None} if no preamble is needed.
        @param sampleIndexFilename: A C{str} filename to write a sample index
            file to. Lines in the file will have an integer index, a space, and
            then the sample name.
        @param pathogenIndexFilename: A C{str} filename to write a pathogen
            index file to. Lines in the file will have an integer index, a
            space, and then the pathogen name.
        @param omitVirusLinks: If C{True}, links to ICTV and ViralZone will be
            omitted in output.
        @param omitSampleProteinCount: If C{True}, do not display a number of
            matched pathogen proteins for a sample. This should be used when
            those numbers are inaccurate (e.g., when using the unclustered RVDB
            protein database and there are many sequences for the same
            protein).
        @return: An HTML C{str} suitable for printing.
        """
        if pathogenType not in ("bacterial", "viral"):
            raise ValueError(
                "Unrecognized pathogenType argument: %r. Value must be either "
                "'bacterial' or 'viral'." % pathogenType
            )

        if not exists(self._pathogenDataDir):
            os.mkdir(self._pathogenDataDir)

        self._computeUniqueReadCounts()

        if pathogenPanelFilename:
            self.pathogenPanel(pathogenPanelFilename)

        if sampleIndexFilename:
            with open(sampleIndexFilename, "w") as fp:
                self.pathogenSampleFiles.writeSampleIndex(fp)

        if pathogenIndexFilename:
            with open(pathogenIndexFilename, "w") as fp:
                self.pathogenSampleFiles.writePathogenIndex(fp)

        # Figure out if we have to delete some pathogens because the number
        # or fraction of its proteins that we have matches for is too low.
        if minProteinFraction > 0.0 or minProteinCount > 0:
            toDelete = defaultdict(list)
            for genomeAccession in self.genomeAccessions:
                genomeInfo = self._db.findGenome(genomeAccession)
                pathogenProteinCount = genomeInfo["proteinCount"]
                assert pathogenProteinCount > 0
                for s in self.genomeAccessions[genomeAccession]:
                    sampleProteinCount = len(
                        self.genomeAccessions[genomeAccession][s]["proteins"]
                    )
                    if sampleProteinCount < minProteinCount:
                        toDelete[genomeAccession].append(s)
                    else:
                        sampleProteinFraction = (
                            sampleProteinCount / pathogenProteinCount
                        )
                        if sampleProteinFraction < minProteinFraction:
                            toDelete[genomeAccession].append(s)

            for genomeAccession, samples in toDelete.items():
                for sample in samples:
                    del self.genomeAccessions[genomeAccession][sample]

        pathogenNames = sorted(
            pathogenName
            for pathogenName in self.pathogenNames
            if len(self.pathogenNames[pathogenName]) > 0
        )
        nPathogenNames = len(pathogenNames)
        sampleNames = sorted(self.sampleNames)

        result = [
            "<html>",
            "<head>",
            "<title>",
            title,
            "</title>",
            '<meta charset="UTF-8">',
            "</head>",
            "<body>",
            "<style>",
            """\
            body {
                margin-left: 2%;
                margin-right: 2%;
            }
            hr {
                display: block;
                margin-top: 0.5em;
                margin-bottom: 0.5em;
                margin-left: auto;
                margin-right: auto;
                border-style: inset;
                border-width: 1px;
            }
            p.pathogen {
                margin-top: 10px;
                margin-bottom: 3px;
            }
            p.sample {
                margin-top: 10px;
                margin-bottom: 3px;
            }
            .sample {
                margin-top: 5px;
                margin-bottom: 2px;
            }
            ul {
                margin-bottom: 2px;
            }
            .indented {
                margin-left: 2em;
            }
            .sample-name {
                font-size: 125%;
                font-weight: bold;
            }
            .pathogen-name {
                font-size: 125%;
                font-weight: bold;
            }
            .index-name {
                font-weight: bold;
            }
            .index {
                font-size: small;
            }
            .protein-name {
                font-family: "Courier New", Courier, monospace;
            }
            .stats {
                font-family: "Courier New", Courier, monospace;
                white-space: pre;
            }
            .protein-list {
                margin-top: 2px;
            }""",
            "</style>",
            "</head>",
            "<body>",
        ]

        proteinFieldsDescription = [
            "<p>",
            "In all bullet point protein lists below, there are the following fields:",
            "<ol>",
            "<li>Coverage fraction.</li>",
            "<li>Median bit score.</li>",
            "<li>Best bit score.</li>",
            "<li>Read count (if the HSP count differs, read and HSP ",
            (
                'counts are both given, separated by "%s").</li>'
                % self.READ_AND_HSP_COUNT_STR_SEP
            ),
            "<li>Protein length (in amino acids).</li>",
        ]

        if self._saveReadLengths:
            proteinFieldsDescription.append(
                "<li>All read lengths (in parentheses).</li>"
            )

        proteinFieldsDescription.extend(
            [
                "<li>Protein name.</li>",
                "</ol>",
                "</p>",
            ]
        )

        append = result.append

        append("<h1>%s</h1>" % title)
        if preamble:
            append("<p>%s</p>" % preamble)
        append("<p>")
        append(self._title())

        if self._pathogenProteinCount and minProteinFraction:
            percent = minProteinFraction * 100.0
            if nPathogenNames < len(self.pathogenNames):
                if nPathogenNames == 1:
                    append(
                        "Pathogen protein fraction filtering has been "
                        "applied, so information on only 1 pathogen is "
                        "displayed. This is the only pathogen for which at "
                        "least one sample matches at least %.2f%% of the "
                        "pathogen proteins." % percent
                    )
                else:
                    append(
                        "Pathogen protein fraction filtering has been "
                        "applied, so information on only %d pathogens is "
                        "displayed. These are the only pathogens for which "
                        "at least one sample matches at least %.2f%% of "
                        "the pathogen proteins." % (nPathogenNames, percent)
                    )
            else:
                append(
                    "Pathogen protein fraction filtering was applied, "
                    "but all pathogens have at least %.2f%% of their "
                    "proteins matched by at least one sample." % percent
                )

        append("</p>")

        if pathogenPanelFilename:
            append("<p>")
            append(
                '<a href="%s">Panel showing read count per pathogen, per '
                "sample.</a>" % pathogenPanelFilename
            )
            append(
                "Red vertical bars indicate samples with an unusually high read count."
            )
            append("</p>")

        result.extend(proteinFieldsDescription)

        # Write a linked table of contents by pathogen.
        append('<p><span class="index-name">Pathogen index:</span>')
        append('<span class="index">')
        for pathogenName in pathogenNames:
            append('<a href="#pathogen-%s">%s</a>' % (pathogenName, pathogenName))
            append("&middot;")
        # Get rid of final middle dot and add a period.
        result.pop()
        result[-1] += "."
        append("</span></p>")

        # Write a linked table of contents by sample.
        append('<p><span class="index-name">Sample index:</span>')
        append('<span class="index">')
        for sampleName in sampleNames:
            append('<a href="#sample-%s">%s</a>' % (sampleName, sampleName))
            append("&middot;")
        # Get rid of final middle dot and add a period.
        result.pop()
        result[-1] += "."
        append("</span></p>")

        # Write all pathogens (with samples (with proteins)).
        append("<hr>")
        append("<h1>Pathogens by sample</h1>")

        for pathogenName in pathogenNames:
            samples = self.pathogenNames[pathogenName]
            sampleCount = len(samples)
            pathogenProteinCount = self._pathogenProteinCount[pathogenName]
            if pathogenType == "viral" and not omitVirusLinks:
                quoted = quote(pathogenName)
                pathogenLinksHTML = (
                    ' (<a href="%s%s">ICTV</a>, <a href="%s%s">ViralZone</a>)'
                ) % (self.ICTV, quoted, self.VIRALZONE, quoted)
            else:
                pathogenLinksHTML = ""

            if pathogenProteinCount:
                withStr = " with %d protein%s" % (
                    pathogenProteinCount,
                    "" if pathogenProteinCount == 1 else "s",
                )
            else:
                withStr = ""

            pathogenIndex = self.pathogenSampleFiles.pathogenIndex(pathogenName)

            pathogenReadsFilename = join(
                self._pathogenDataDir, "pathogen-%d.%s" % (pathogenIndex, self._format)
            )

            pathogenReadsFp = open(pathogenReadsFilename, "w")
            pathogenReadCount = 0

            append(
                '<a id="pathogen-%s"></a>'
                '<p class="pathogen">'
                '<span class="pathogen-name">%s</span>'
                "%s %s, "
                "was matched by %d sample%s "
                '(<a href="%s">%s</a> in total):'
                "</p>"
                % (
                    pathogenName,
                    pathogenName,
                    pathogenLinksHTML,
                    withStr,
                    sampleCount,
                    "" if sampleCount == 1 else "s",
                    pathogenReadsFilename,
                    self.READCOUNT_MARKER,
                )
            )

            # Remember where we are in the output result so we can fill in
            # the total read count once we have processed all samples for
            # this pathogen. Not nice, I know.
            pathogenReadCountLineIndex = len(result) - 1

            for sampleName in sorted(samples):
                readsFileName = self.pathogenSampleFiles.lookup(
                    pathogenName, sampleName
                )

                # Copy the read data from the per-sample reads for this
                # pathogen into the per-pathogen file of reads.
                with open(readsFileName) as readsFp:
                    while True:
                        data = readsFp.read(4096)
                        if data:
                            pathogenReadsFp.write(data)
                        else:
                            break

                proteins = samples[sampleName]["proteins"]
                proteinCount = len(proteins)
                uniqueReadCount = samples[sampleName]["uniqueReadCount"]
                pathogenReadCount += uniqueReadCount

                if omitSampleProteinCount:
                    proteinCountHTML = ""
                else:
                    proteinCountHTML = "%d protein%s, " % (
                        proteinCount,
                        "" if proteinCount == 1 else "s",
                    )

                append(
                    '<p class="sample indented">'
                    'Sample <a href="#sample-%s">%s</a> '
                    '(%s<a href="%s">%d de-duplicated (by id) '
                    'read%s</a>, <a href="%s">panel</a>):</p>'
                    % (
                        sampleName,
                        sampleName,
                        proteinCountHTML,
                        readsFileName,
                        uniqueReadCount,
                        "" if uniqueReadCount == 1 else "s",
                        self.sampleNames[sampleName],
                    )
                )
                append('<ul class="protein-list indented">')
                for proteinName in sorted(proteins):
                    proteinMatch = proteins[proteinName]
                    append(
                        "<li>"
                        '<span class="stats">'
                        "%(coverage).2f %(medianScore)6.2f %(bestScore)6.2f "
                        "%(readAndHspCountStr)11s %(proteinLength)4d " % proteinMatch
                    )

                    if self._saveReadLengths:
                        append(
                            "(%s) "
                            % ", ".join(map(str, sorted(proteinMatch["readLengths"])))
                        )

                    append(
                        "</span> "
                        '<span class="protein-name">'
                        "%(proteinName)s"
                        "</span> "
                        '(<a href="%(bluePlotFilename)s">blue plot</a>, '
                        '<a href="%(readsFilename)s">reads</a>' % proteinMatch
                    )

                    if proteinMatch["proteinURL"]:
                        # Append this directly to the last string in result, to
                        # avoid introducing whitespace when we join result
                        # using '\n'.
                        result[-1] += (
                            ', <a href="%s">NCBI protein</a>'
                            % proteinMatch["proteinURL"]
                        )

                    if proteinMatch["genomeURL"]:
                        # Append this directly to the last string in result, to
                        # avoid introducing whitespace when we join result
                        # using '\n'.
                        result[-1] += (
                            ', <a href="%s">NCBI genome</a>' % proteinMatch["genomeURL"]
                        )
                    result[-1] += ")"

                    append("</li>")

                append("</ul>")

            pathogenReadsFp.close()

            # Sanity check there's a read count marker text in our output
            # where we expect it.
            readCountLine = result[pathogenReadCountLineIndex]
            if readCountLine.find(self.READCOUNT_MARKER) == -1:
                raise ValueError(
                    "Could not find pathogen read count marker (%s) in result "
                    "index %d text (%s)."
                    % (self.READCOUNT_MARKER, pathogenReadCountLineIndex, readCountLine)
                )

            # Put the read count into the pathogen summary line we wrote
            # earlier, replacing the read count marker with the correct
            # text.
            result[pathogenReadCountLineIndex] = readCountLine.replace(
                self.READCOUNT_MARKER,
                "%d read%s"
                % (pathogenReadCount, "" if pathogenReadCount == 1 else "s"),
            )

        # Write all samples (with pathogens (with proteins)).
        append("<hr>")
        append("<h1>Samples by pathogen</h1>")

        for sampleName in sampleNames:
            samplePathogenNames = [
                pathName
                for pathName in self.pathogenNames
                if sampleName in self.pathogenNames[pathName]
            ]

            if len(samplePathogenNames):
                append(
                    '<a id="sample-%s"></a>'
                    '<p class="sample">Sample '
                    '<span class="sample-name">%s</span> '
                    "matched proteins from %d pathogen%s, "
                    '<a href="%s">panel</a>:</p>'
                    % (
                        sampleName,
                        sampleName,
                        len(samplePathogenNames),
                        "" if len(samplePathogenNames) == 1 else "s",
                        self.sampleNames[sampleName],
                    )
                )
            else:
                append(
                    '<a id="sample-%s"></a>'
                    '<p class="sample">Sample '
                    '<span class="sample-name">%s</span> '
                    "did not match anything.</p>" % (sampleName, sampleName)
                )
                continue

            for pathogenName in sorted(samplePathogenNames):
                readsFileName = self.pathogenSampleFiles.lookup(
                    pathogenName, sampleName
                )
                proteins = self.pathogenNames[pathogenName][sampleName]["proteins"]
                uniqueReadCount = self.pathogenNames[pathogenName][sampleName][
                    "uniqueReadCount"
                ]
                proteinCount = len(proteins)
                pathogenProteinCount = self._pathogenProteinCount[pathogenName]

                if pathogenProteinCount:
                    proteinCountStr = "%d/%d protein%s" % (
                        proteinCount,
                        pathogenProteinCount,
                        "" if pathogenProteinCount == 1 else "s",
                    )
                else:
                    proteinCountStr = "%d protein%s" % (
                        proteinCount,
                        "" if proteinCount == 1 else "s",
                    )

                append(
                    '<p class="sample indented">'
                    '<a href="#pathogen-%s">%s</a> %s, '
                    '<a href="%s">%d de-duplicated (by id) read%s</a>:</p>'
                    % (
                        pathogenName,
                        pathogenName,
                        proteinCountStr,
                        readsFileName,
                        uniqueReadCount,
                        "" if uniqueReadCount == 1 else "s",
                    )
                )
                append('<ul class="protein-list indented">')
                for proteinName in sorted(proteins):
                    proteinMatch = proteins[proteinName]
                    append(
                        "<li>"
                        '<span class="stats">'
                        "%(coverage).2f %(medianScore)6.2f %(bestScore)6.2f "
                        "%(readAndHspCountStr)11s %(proteinLength)4d "
                        "</span> "
                        '<span class="protein-name">'
                        "%(proteinName)s"
                        "</span> "
                        '(<a href="%(bluePlotFilename)s">blue plot</a>, '
                        '<a href="%(readsFilename)s">reads</a>' % proteinMatch
                    )

                    if proteinMatch["proteinURL"]:
                        # Append this directly to the last string in result, to
                        # avoid introducing whitespace when we join result
                        # using '\n'.
                        result[-1] += (
                            ', <a href="%s">NCBI protein</a>'
                            % proteinMatch["proteinURL"]
                        )

                    if proteinMatch["genomeURL"]:
                        # Append this directly to the last string in result, to
                        # avoid introducing whitespace when we join result
                        # using '\n'.
                        result[-1] += (
                            ', <a href="%s">NCBI genome</a>' % proteinMatch["genomeURL"]
                        )

                    result[-1] += ")"

                    append("</li>")

                append("</ul>")

        append("</body>")
        append("</html>")

        return "\n".join(result)

    def _pathogenSamplePlot(self, pathogenName, sampleNames, ax):
        """
        Make an image of a graph giving pathogen read count (Y axis) versus
        sample id (X axis).

        @param pathogenName: A C{str} pathogen name.
        @param sampleNames: A sorted C{list} of sample names.
        @param ax: A matplotlib C{axes} instance.
        """
        readCounts = []
        for i, sampleName in enumerate(sampleNames):
            try:
                readCount = self.pathogenNames[pathogenName][sampleName][
                    "uniqueReadCount"
                ]
            except KeyError:
                readCount = 0
            readCounts.append(readCount)

        highlight = "r"
        normal = "gray"
        sdMultiple = 2.5
        minReadsForHighlighting = 10
        highlighted = []

        if len(readCounts) == 1:
            if readCounts[0] > minReadsForHighlighting:
                color = [highlight]
                highlighted.append(sampleNames[0])
            else:
                color = [normal]
        else:
            mean = np.mean(readCounts)
            sd = np.std(readCounts)
            color = []
            for readCount, sampleName in zip(readCounts, sampleNames):
                if (
                    readCount > (sdMultiple * sd) + mean
                    and readCount >= minReadsForHighlighting
                ):
                    color.append(highlight)
                    highlighted.append(sampleName)
                else:
                    color.append(normal)

        nSamples = len(sampleNames)
        x = np.arange(nSamples)
        yMin = np.zeros(nSamples)
        ax.set_xticks([])
        ax.set_xlim((-0.5, nSamples - 0.5))
        ax.vlines(x, yMin, readCounts, color=color)
        if highlighted:
            title = "%s\nIn red: %s" % (pathogenName, fill(", ".join(highlighted), 50))
        else:
            # Add a newline to keep the first line of each title at the
            # same place as those titles that have an "In red:" second
            # line.
            title = pathogenName + "\n"

        ax.set_title(title, fontsize=10)
        ax.tick_params(axis="both", which="major", labelsize=8)
        ax.tick_params(axis="both", which="minor", labelsize=6)

    def pathogenPanel(self, filename):
        """
        Make a panel of images, with each image being a graph giving pathogen
        de-duplicated (by id) read count (Y axis) versus sample id (X axis).

        @param filename: A C{str} file name to write the image to.
        """
        import matplotlib

        matplotlib.use("PDF")
        import matplotlib.pyplot as plt

        self._computeUniqueReadCounts()
        pathogenNames = sorted(self.pathogenNames)
        sampleNames = sorted(self.sampleNames)

        cols = 5
        rows = int(len(pathogenNames) / cols) + (
            0 if len(pathogenNames) % cols == 0 else 1
        )
        figure, ax = plt.subplots(rows, cols, squeeze=False)

        coords = dimensionalIterator((rows, cols))

        for i, pathogenName in enumerate(pathogenNames):
            row, col = next(coords)
            self._pathogenSamplePlot(pathogenName, sampleNames, ax[row][col])

        # Hide the final panel graphs (if any) that have no content. We do
        # this because the panel is a rectangular grid and some of the
        # plots at the end of the last row may be unused.
        for row, col in coords:
            ax[row][col].axis("off")

        figure.suptitle(
            (
                "Per-sample read count for %d pathogen%s and %d sample%s.\n\n"
                "Sample name%s: %s"
            )
            % (
                len(pathogenNames),
                "" if len(pathogenNames) == 1 else "s",
                len(sampleNames),
                "" if len(sampleNames) == 1 else "s",
                "" if len(sampleNames) == 1 else "s",
                fill(", ".join(sampleNames), 50),
            ),
            fontsize=20,
        )
        figure.set_size_inches(5.0 * cols, 2.0 * rows, forward=True)
        plt.subplots_adjust(hspace=0.4)

        figure.savefig(filename)



import sqlite3
from unittest import TestCase

from dark.taxonomy import (
    LineageElement as LE,
)
from dark.taxonomy import (
    LineageFetcher,
    Taxonomy,
    isAllowedTaxonomicRank,
    isRetrovirus,
    isRNAVirus,
)


class FakeCursor:
    def __init__(self, results):
        self._results = results
        self._index = -1

    def execute(self, p):
        pass

    def fetchone(self):
        self._index += 1
        return self._results[self._index]

    def close(self):
        pass


class FakeDbConnection:
    def __init__(self, results):
        self._results = results
        self.open = True

    def cursor(self):
        return FakeCursor(self._results)

    def close(self):
        self.open = False


class TestLineageFetcher(TestCase):
    """
    Test LineageFetcher class.
    """

    def testGetTaxonomy(self):
        """
        Test if the LineageFetcher class works properly.
        """
        title = "gi|5|gb|EU375804.1| Merkel cell polyomavirus"

        db = FakeDbConnection(
            [
                [15],
                ["Merkel cell polyomavirus"],
                [4],
                ["Polyomavirus"],
                [3],
                ["dsDNA viruses"],
                [2],
                ["Vira"],
                [1],
            ]
        )
        cursor = db.cursor()

        lineageFetcher = LineageFetcher(db=db, cursor=cursor)

        lineage = lineageFetcher.lineage(title)
        self.assertEqual(
            [
                (15, "Merkel cell polyomavirus"),
                (4, "Polyomavirus"),
                (3, "dsDNA viruses"),
                (2, "Vira"),
            ],
            lineage,
        )


class TestTaxonomy(TestCase):
    """
    Test the Taxonomy class.
    """

    def _makeFetcher(self, data):
        """
        Make a taxonomy database and put it into an Taxonomy
        instance.

        @param data: A C{dict} with 'taxids', 'names', and 'nodes' keys,
            each of which is a C{list} of C{tuple}s containing values to
            insert into the respective tables.
        @return: A C{Taxonomy} instance.
        """
        db = sqlite3.connect(":memory:")
        cursor = db.cursor()

        # The tables created here must be identical to those expected by
        # the Taxonomy class. See also
        # https://github.com/acorg/ncbi-taxonomy-database
        cursor.executescript(
            """
            CREATE TABLE nodes (
                taxid INTEGER NOT NULL,
                parent_taxid INTEGER NOT NULL,
                rank VARCHAR NOT NULL
            );

            CREATE TABLE accession_taxid (
                accession VARCHAR UNIQUE PRIMARY KEY,
                taxid INTEGER NOT NULL
            );

            CREATE TABLE names (
                taxid INTEGER NOT NULL,
                name VARCHAR NOT NULL
            );

            CREATE INDEX nodes_idx ON nodes(taxid);
            CREATE INDEX accession_idx ON accession_taxid(accession);
            CREATE INDEX name_idx ON names(taxid);
        """
        )

        execute = cursor.execute

        for accession, taxid in data.get("taxids", []):
            execute("INSERT INTO accession_taxid VALUES (?, ?)", (accession, taxid))

        for taxid, name in data.get("names", []):
            execute("INSERT INTO names VALUES (?, ?)", (taxid, name))

        for taxid, parentTaxid, rank in data.get("nodes", []):
            execute("INSERT INTO nodes VALUES (?, ?, ?)", (taxid, parentTaxid, rank))

        return Taxonomy(db)

    def testUnknownTaxid(self):
        """
        If a taxonomy id is not present in the accession_taxid table, the
        lineage fetcher must raise a C{ValueError}.
        """
        fetcher = self._makeFetcher({})
        error = r"^Could not find 'DQ011818\.1' in accession_taxid or names " r"tables$"
        self.assertRaisesRegex(ValueError, error, fetcher.lineage, "DQ011818.1")
        fetcher.close()

    def testReusingClosedFetcher(self):
        """
        Trying to re-use a closed fetcher must result in an AttributeError
        being raised.
        """
        fetcher = self._makeFetcher({})
        fetcher.close()
        error = "^'NoneType' object has no attribute 'cursor'$"
        self.assertRaisesRegex(AttributeError, error, fetcher.lineage, "DQ011818.1")

    def testTaxidNotInNamesTable(self):
        """
        If a taxonomy id is not present in the names table, a ValueError must
        be raised.
        """
        fetcher = self._makeFetcher(
            {
                "taxids": (("DQ011818.1", 500),),
            }
        )
        error = "^Could not find taxonomy id 500 in names table$"
        self.assertRaisesRegex(ValueError, error, fetcher.lineage, "DQ011818.1")
        fetcher.close()

    def testTaxidNotInNodesTable(self):
        """
        If a taxonomy id is not present in the nodes table, a ValueError must
        be raised.
        """
        fetcher = self._makeFetcher(
            {
                "taxids": (("DQ011818.1", 500),),
                "names": ((500, "Porcine endogenous retrovirus A"),),
            }
        )
        error = "^Could not find taxonomy id 500 in nodes table$"
        self.assertRaisesRegex(ValueError, error, fetcher.lineage, "DQ011818.1")
        fetcher.close()

    def testLookupByAccession(self):
        """
        The full taxonomy of an accession number must be retrievable.
        """
        fetcher = self._makeFetcher(
            {
                "taxids": (("DQ011818.1", 500),),
                "names": (
                    (500, "Porcine endogenous retrovirus A"),
                    (400, "Retroviruses"),
                    (300, "Viruses"),
                ),
                "nodes": (
                    (500, 400, "species"),
                    (400, 300, "genus"),
                    (300, 1, "realm"),
                ),
            }
        )

        self.assertEqual(
            (
                (500, "Porcine endogenous retrovirus A", "species"),
                (400, "Retroviruses", "genus"),
                (300, "Viruses", "realm"),
            ),
            fetcher.lineage("DQ011818.1"),
        )
        fetcher.close()

    def testLookupByName(self):
        """
        The taxonomy must be retrievable by name.
        """
        fetcher = self._makeFetcher(
            {
                "taxids": (("DQ011818.1", 500),),
                "names": (
                    (500, "Porcine endogenous retrovirus A"),
                    (400, "Retroviruses"),
                    (300, "Viruses"),
                ),
                "nodes": (
                    (500, 400, "species"),
                    (400, 300, "genus"),
                    (300, 1, "realm"),
                ),
            }
        )

        self.assertEqual(
            (
                (500, "Porcine endogenous retrovirus A", "species"),
                (400, "Retroviruses", "genus"),
                (300, "Viruses", "realm"),
            ),
            fetcher.lineage("Porcine endogenous retrovirus A"),
        )
        fetcher.close()

    def testLookupByTaxid(self):
        """
        The taxonomy must be retrievable by taxonomy id.
        """
        fetcher = self._makeFetcher(
            {
                "taxids": (("DQ011818.1", 500),),
                "names": (
                    (500, "Porcine endogenous retrovirus A"),
                    (400, "Retroviruses"),
                    (300, "Viruses"),
                ),
                "nodes": (
                    (500, 400, "species"),
                    (400, 300, "genus"),
                    (300, 1, "realm"),
                ),
            }
        )

        self.assertEqual(
            (
                (500, "Porcine endogenous retrovirus A", "species"),
                (400, "Retroviruses", "genus"),
                (300, "Viruses", "realm"),
            ),
            fetcher.lineage(500),
        )
        fetcher.close()

    def testLookupWithSkip(self):
        """
        It must be possible to skip a level when retrieving a taxonomy.
        """

        def skipFunc(lineageElement):
            return lineageElement.rank == "genus"

        fetcher = self._makeFetcher(
            {
                "taxids": (("DQ011818.1", 500),),
                "names": (
                    (500, "Porcine endogenous retrovirus A"),
                    (400, "Retroviruses"),
                    (300, "Viruses"),
                ),
                "nodes": (
                    (500, 400, "species"),
                    (400, 300, "genus"),
                    (300, 1, "realm"),
                ),
            }
        )

        self.assertEqual(
            (
                (500, "Porcine endogenous retrovirus A", "species"),
                (300, "Viruses", "realm"),
            ),
            fetcher.lineage(500, skipFunc=skipFunc),
        )
        fetcher.close()

    def testLookupWithStop(self):
        """
        It must be possible to stop at a level when retrieving a taxonomy.
        """

        def stopFunc(lineageElement):
            return lineageElement.rank == "genus"

        fetcher = self._makeFetcher(
            {
                "taxids": (("DQ011818.1", 500),),
                "names": (
                    (500, "Porcine endogenous retrovirus A"),
                    (400, "Retroviruses"),
                    (300, "Viruses"),
                ),
                "nodes": (
                    (500, 400, "species"),
                    (400, 300, "genus"),
                    (300, 1, "realm"),
                ),
            }
        )

        self.assertEqual(
            (
                (500, "Porcine endogenous retrovirus A", "species"),
                (400, "Retroviruses", "genus"),
            ),
            fetcher.lineage(500, stopFunc=stopFunc),
        )
        fetcher.close()

    def testLookupWithSkipAndStop(self):
        """
        It must be possible to stop at a level when retrieving a taxonomy
        and skip the last (stopping) level.
        """

        def skipFunc(lineageElement):
            return lineageElement.rank == "genus"

        def stopFunc(lineageElement):
            return lineageElement.rank == "genus"

        fetcher = self._makeFetcher(
            {
                "taxids": (("DQ011818.1", 500),),
                "names": (
                    (500, "Porcine endogenous retrovirus A"),
                    (400, "Retroviruses"),
                    (300, "Viruses"),
                ),
                "nodes": (
                    (500, 400, "species"),
                    (400, 300, "genus"),
                    (300, 1, "realm"),
                ),
            }
        )

        self.assertEqual(
            ((500, "Porcine endogenous retrovirus A", "species"),),
            fetcher.lineage(500, skipFunc=skipFunc, stopFunc=stopFunc),
        )
        fetcher.close()

    def testSubsetLineageByRanks(self):
        """
        Test the Taxonomy.subsetLineageByRanks static method.
        """
        lineage = (
            LE(11234, "Measles morbillivirus", "species"),
            LE(11229, "Morbillivirus", "genus"),
            LE(2560076, "Orthoparamyxovirinae", "subfamily"),
            LE(11158, "Paramyxoviridae", "family"),
            LE(11157, "Mononegavirales", "order"),
            LE(2497574, "Monjiviricetes", "class"),
            LE(2497570, "Haploviricotina", "subphylum"),
            LE(2497569, "Negarnaviricota", "phylum"),
            LE(2559587, "Riboviria", "realm"),
            LE(10239, "Viruses", "superkingdom"),
        )

        def filterFunc(rank):
            return rank in {"phylum", "genus"}

        self.assertEqual(
            (
                LE(11229, "Morbillivirus", "genus"),
                LE(2497569, "Negarnaviricota", "phylum"),
            ),
            tuple(Taxonomy.subsetLineageByRanks(lineage, filterFunc)),
        )


class TestIsRetrovirus(TestCase):
    """
    Test the isRetrovirus function.
    """

    def testYes(self):
        """
        A retrovirus must be detected.
        """
        self.assertTrue(
            isRetrovirus(
                (
                    LE(11676, "Human immunodeficiency virus 1", "species"),
                    LE(11646, "Lentivirus", "genus"),
                    LE(327045, "Orthoretrovirinae", "subfamily"),
                    LE(11632, "Retroviridae", "family"),
                    LE(2169561, "Ortervirales", "order"),
                    LE(10239, "Viruses", "superkingdom"),
                )
            )
        )

    def testNo(self):
        """
        A non-retrovirus must be detected.
        """
        self.assertFalse(
            isRetrovirus(
                (
                    LE(10310, "Human alphaherpesvirus 2", "species"),
                    LE(10294, "Simplexvirus", "genus"),
                    LE(10293, "Alphaherpesvirinae", "subfamily"),
                    LE(10292, "Herpesviridae", "family"),
                    LE(548681, "Herpesvirales", "order"),
                    LE(10239, "Viruses", "superkingdom"),
                )
            )
        )


class TestIsRNAVirus(TestCase):
    """
    Test the isRNAVirus function.
    """

    def testByRealm(self):
        """
        Anything in the Riboviria realm must be classified as an RNA virus.
        """
        self.assertTrue(
            isRNAVirus(
                (
                    LE(11234, "Measles morbillivirus", "species"),
                    LE(11229, "Morbillivirus", "genus"),
                    LE(2560076, "Orthoparamyxovirinae", "subfamily"),
                    LE(11158, "Paramyxoviridae", "family"),
                    LE(11157, "Mononegavirales", "order"),
                    LE(2497574, "Monjiviricetes", "class"),
                    LE(2497570, "Haploviricotina", "subphylum"),
                    LE(2497569, "Negarnaviricota", "phylum"),
                    LE(2559587, "Riboviria", "realm"),
                    LE(10239, "Viruses", "superkingdom"),
                )
            )
        )

    def testHBV(self):
        """
        An HBV virus must not be classified as an RNA virus.
        """
        self.assertFalse(
            isRNAVirus(
                (
                    LE(1508712, "Tent-making bat hepatitis B virus", "species"),
                    LE(10405, "Orthohepadnavirus", "genus"),
                    LE(10404, "Hepadnaviridae", "family"),
                    LE(10239, "Viruses", "superkingdom"),
                )
            )
        )

    def testHIV(self):
        """
        An HIV virus must be classified as an RNA virus.
        """
        self.assertTrue(
            isRNAVirus(
                (
                    LE(11676, "Human immunodeficiency virus 1", "species"),
                    LE(11646, "Lentivirus", "genus"),
                    LE(327045, "Orthoretrovirinae", "subfamily"),
                    LE(11632, "Retroviridae", "family"),
                    LE(2169561, "Ortervirales", "order"),
                    LE(10239, "Viruses", "superkingdom"),
                )
            )
        )


class TestAllowedTaxonomicRank(TestCase):
    """
    Test the isAllowedTaxonomicRank function.
    """

    def testNoRanksAllowed(self):
        """
        The function must return False when no taxonomic ranks are allowed.
        """
        allowed = set()
        lineage = (
            LE(245, "x", "realm"),
            LE(246, "bacteria", "kingdom"),
            LE(247, "y", "species"),
        )
        self.assertFalse(isAllowedTaxonomicRank(allowed, lineage))

    def testEmptyLineage(self):
        """
        The function must return False when the passed lineage is empty.
        """
        allowed = set((("bacteria", "kingdom"),))
        lineage = ()
        self.assertFalse(isAllowedTaxonomicRank(allowed, lineage))

    def testExactCase(self):
        """
        The function must return True when case matches exactly.
        """
        allowed = set((("bacteria", "kingdom"),))
        lineage = (
            LE(245, "x", "realm"),
            LE(246, "bacteria", "kingdom"),
            LE(247, "y", "species"),
        )
        self.assertTrue(isAllowedTaxonomicRank(allowed, lineage))

    def testNonMatchingCase(self):
        """
        The function must return True when case doesn't match.
        """
        allowed = set((("BACTERIA", "KINGDOM"),))
        lineage = (
            LE(245, "x", "realm"),
            LE(246, "bacteria", "kingdom"),
            LE(247, "y", "species"),
        )
        self.assertTrue(isAllowedTaxonomicRank(allowed, lineage))

    def testMultipleMatches(self):
        """
        The function must return True when more than one part of the lineage matches.
        """
        allowed = set((("bacteria", "kingdom"), ("y", "species")))
        lineage = (
            LE(245, "x", "realm"),
            LE(246, "bacteria", "kingdom"),
            LE(247, "y", "species"),
        )
        self.assertTrue(isAllowedTaxonomicRank(allowed, lineage))

    def testRankMismatch(self):
        """
        The function must return False when the name matches but the rank does not.
        """
        allowed = set((("bacteria", "species"),))
        lineage = (
            LE(245, "x", "realm"),
            LE(246, "bacteria", "kingdom"),
            LE(247, "y", "species"),
        )
        self.assertFalse(isAllowedTaxonomicRank(allowed, lineage))

    def testNameMismatch(self):
        """
        The function must return False when the rank matches but the name does not.
        """
        allowed = set((("virus", "kingdom"),))
        lineage = (
            LE(245, "x", "realm"),
            LE(246, "bacteria", "kingdom"),
            LE(247, "y", "species"),
        )
        self.assertFalse(isAllowedTaxonomicRank(allowed, lineage))

    def testNameAndRankMismatch(self):
        """
        The function must return False when the name and rank both do not match.
        """
        allowed = set((("virus", "genus"),))
        lineage = (
            LE(245, "x", "realm"),
            LE(246, "bacteria", "kingdom"),
            LE(247, "y", "species"),
        )
        self.assertFalse(isAllowedTaxonomicRank(allowed, lineage))


class TestLineageElement(TestCase):
    """
    Test the LineageElement named tuple.
    """

    def testTaxid(self):
        """
        The taxid attribute must be set as expected.
        """
        element = LE(245, "no name", "species")
        self.assertEqual(245, element.taxid)
        self.assertEqual(245, element[0])

    def testName(self):
        """
        The name attribute must be set as expected.
        """
        element = LE(245, "no name", "species")
        self.assertEqual("no name", element.name)
        self.assertEqual("no name", element[1])

    def testRank(self):
        """
        The rank attribute must be set as expected.
        """
        element = LE(245, "no name", "species")
        self.assertEqual("species", element.rank)
        self.assertEqual("species", element[2])



#!/usr/bin/env python

import sys

from dark.summarize import summarizeReads

if len(sys.argv) != 2:
    print("Usage: %s file.fasta / file.fastq" % sys.argv[0], file=sys.stderr)
    sys.exit(1)
else:
    filename = sys.argv[1]
    if filename.lower().endswith("q"):
        filetype = "fastq"
    else:
        # Assume anything not ending with a 'q' is fasta.
        filetype = "fasta"

    result = summarizeReads(filename, filetype)

    print("Number of reads:", result["read_number"])
    print("Total length: %s bases" % result["total_length"])
    print("The average read length: %s bases" % result["average_length"])
    print("Longest read:", result["max_length"])
    print("Shortest read:", result["min_length"])
    print("Median length:", result["median_length"])

    for base in sorted(result["base_counts"]):
        count = result["base_counts"][base]
        print(
            "%s: Total: %s; Average per read: %s"
            % (base, count, count / result["read_number"])
        )



#!/usr/bin/env python

import argparse
import os
import sys
from os.path import basename, join
from tempfile import mkdtemp

from dark.fasta import FastaReads
from dark.process import Executor

IVAR_FREQUENCY_THRESHOLD_DEFAULT = 0.6
IVAR_DOCS = "https://andersen-lab.github.io/ivar/html/manualpage.html#autotoc_md19"


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Make a consensus sequence.",
    )

    parser.add_argument("--reference", required=True, help="The reference FASTA file.")

    parser.add_argument(
        "--bam",
        help=(
            "The BAM file from which the consensus should be made. "
            "Required if --maskLowCoverage is used. If no BAM file is "
            "given, a VCF file must be provided. If both a BAM and a VCF "
            "file are given, the VCF file will take precedence."
        ),
    )

    parser.add_argument(
        "--vcfFile",
        help=(
            "The VCF file. If omitted, bcftools will be used to make a VCF "
            "file from the BAM file."
        ),
    )

    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        "--id",
        help=(
            "The id to use in the consensus sequence in the output FASTA. "
            "If not given, the reference sequence id will be used."
        ),
    )

    group.add_argument(
        "--idLambda",
        metavar="LAMBDA-FUNCTION",
        help=(
            "A one-argument function taking and returning a read id. "
            "This can be used to set the id of the reference sequence based "
            "on the id of the reference sequence (the function will be "
            "called with the id of the reference sequence). E.g., "
            "--idLambda \"lambda id: id.split('_')[0]\" or "
            "--idLambda \"lambda id: id[:10] + '-consensus'\"."
        ),
    )

    parser.add_argument(
        "--sample",
        help=(
            "The name of the sample (from the @RG SM tag in the original "
            "alignment BAM file) for which a consensus should be made. "
            "If not given, the first sample name (from the #CHROM header) "
            "in the VCF file will be used."
        ),
    )

    parser.add_argument(
        "--dryRun",
        action="store_true",
        help="Do not run commands, just print what would be done.",
    )

    parser.add_argument(
        "--maskLowCoverage",
        default=0,
        type=int,
        help=(
            "Put an N into sites where the coverage is below the specified "
            "cutoff. If you specify a negative numer, masking will be "
            "turned off. Requires --bam."
        ),
    )

    parser.add_argument(
        "--log",
        action="store_true",
        help=(
            "Show a log of commands that were (or would be, if --dryRun is "
            "used) executed."
        ),
    )

    parser.add_argument(
        "--noClean",
        action="store_false",
        dest="clean",
        help="Do not remove intermediate files or the temporary directory.",
    )

    parser.add_argument(
        "--callHaplotypesGATK",
        action="store_true",
        help=(
            "Use GATK to call haplotypes. See "
            "https://gatk.broadinstitute.org for details on GATK."
        ),
    )

    parser.add_argument(
        "--picardJar",
        help=(
            "The path to the Picard jar file. See "
            "https://github.com/broadinstitute/picard for details on "
            "Picard."
        ),
    )

    parser.add_argument(
        "--ivar",
        action="store_true",
        help="If given, ivar will be used to call the consensus.",
    )

    parser.add_argument(
        "--ivarFrequencyThreshold",
        type=float,
        help=(
            f"The frequency threshold used by ivar when calling the "
            f"consensus. If the frequency of the most-common nucleotide at "
            f"a site meets this threshold, the nucleotide will be called. "
            f"Otherwise, an ambiguous nucleotide code will be produced, "
            f"based on the smallest set of most-frequent nucleotides whose "
            f"summed frequencies meet the threshold. See {IVAR_DOCS} for "
            f"more information. If not given, "
            f"{IVAR_FREQUENCY_THRESHOLD_DEFAULT} is used. Can only be used "
            f"if --ivar is also specified."
        ),
    )

    parser.add_argument(
        "--ivarBedFile",
        help="If ivar should trim primers, a BED file of the primer positions.",
    )

    args = parser.parse_args()

    if not (args.bam or args.vcfFile):
        print("At least one of --bam or --vcfFile must be given.", file=sys.stderr)
        sys.exit(1)

    if args.maskLowCoverage and not args.bam:
        print("If --maskLowCoverage is used, --bam must be too.", file=sys.stderr)
        sys.exit(1)

    if args.ivar and not args.bam:
        print("If --ivar is used, --bam must be too.", file=sys.stderr)
        sys.exit(1)

    if args.ivarFrequencyThreshold is not None and not args.ivar:
        print(
            "If --ivarFrequencyThreshold is used, --ivar must be too.", file=sys.stderr
        )
        sys.exit(1)

    if args.ivar and args.ivarFrequencyThreshold is None:
        args.ivarFrequencyThreshold = IVAR_FREQUENCY_THRESHOLD_DEFAULT

    e = Executor(args.dryRun)

    tempdir = mkdtemp(prefix="consensus-")

    if args.vcfFile:
        vcfFile = args.vcfFile
    else:
        # No VCF file provided, so make one.
        vcfFile = join(tempdir, "vcf.gz")
        if args.callHaplotypesGATK:
            e.execute(f"samtools index {args.bam!r}")
            if args.picardJar:
                picardJar = args.picardJar
            else:
                try:
                    picardJar = os.environ["PICARD_JAR"]
                except KeyError:
                    print(
                        "If you use --callHaplotypesGATK, you must give a "
                        "Picard JAR file with --picardJar or else set "
                        "PICARD_JAR in your environment.",
                        file=sys.stderr,
                    )
                    sys.exit(1)

            indexFile = args.reference + ".fai"
            if os.path.exists(indexFile):
                removeIndex = False
            else:
                removeIndex = True
                e.execute(f"samtools faidx {args.reference!r}")

            if args.reference.lower().endswith(".fasta"):
                dictFile = args.reference[: -len(".fasta")] + ".dict"
            else:
                dictFile = args.reference + ".dict"

            if os.path.exists(dictFile):
                removeDict = False
            else:
                removeDict = True
                e.execute(
                    f"java -jar {picardJar!r} CreateSequenceDictionary "
                    f"R={args.reference!r} O={dictFile!r}"
                )

            e.execute(
                f"gatk --java-options -Xmx4g HaplotypeCaller "
                f"--reference {args.reference!r} "
                f"--input {args.bam!r} "
                f"--output {vcfFile!r} "
                f"--sample-ploidy 1 "
                f"-ERC GVCF"
            )

            if removeIndex:
                e.execute(f"rm {indexFile!r}")

            if removeDict:
                e.execute(f"rm {dictFile!r}")
        else:
            e.execute(
                f"bcftools mpileup --max-depth 5000 -Ou -f {args.reference!r} "
                f"{args.bam!r} | bcftools call --ploidy 1 -mv -Oz -o {vcfFile!r}"
            )

            e.execute(f"bcftools index {vcfFile!r}")

    if args.maskLowCoverage >= 0:
        # Make a BED file.
        bedFile = join(tempdir, "mask.bed")
        # The doubled-% below are so that Python doesn't try to fill in the
        # values and instead just generates a single % that awk sees.
        e.execute(
            f"samtools depth -a {args.bam!r} | "
            f"awk '$3 < {args.maskLowCoverage} "
            f'{{printf "%s\\t%d\\t%d\\n", $1, $2 - 1, $2}}\' > {bedFile!r}'
        )
        maskArg = "--mask " + bedFile
    else:
        maskArg = ""

    if args.sample:
        sample = args.sample
    else:
        result = e.execute(f"gunzip -c {vcfFile!r} | egrep -m 1 '^#CHROM' | cut -f10")
        sample = "SAMPLE-NAME" if args.dryRun else result.stdout.strip()

    consensusFile = join(tempdir, "consensus.fasta")

    if args.ivar:
        if args.ivarBedFile:
            tempBamFile = join(tempdir, basename(args.bam) + "-trimmed")
            result = e.execute(
                f"ivar trim -i {args.bam!r} -b {args.ivarBedFile!r} -p {tempBamFile!r} "
                "-e"
            )
            ivarTempBamFile = tempBamFile + ".bam"
            sortedIvarTempBamFile = tempBamFile + "-trimmed-sorted.bam"
            result = e.execute(
                f"samtools sort {ivarTempBamFile!r} -o {sortedIvarTempBamFile!r}"
            )
            bamFile = sortedIvarTempBamFile
        else:
            bamFile = args.bam

        ivarConsensusFile = join(tempdir, "temporary-consensus")
        result = e.execute(
            f"samtools mpileup -A -Q 0 {bamFile!r} | "
            f"ivar consensus -p {ivarConsensusFile!r} -q 20 "
            f"-t {args.ivarFrequencyThreshold!r} -m {args.maskLowCoverage!r}"
        )

        result = e.execute(f"mv {(ivarConsensusFile + '.fa')!r} {consensusFile!r}")

    else:
        result = e.execute(
            f"bcftools consensus --sample {sample!r} --iupac-codes {maskArg} "
            f"--fasta-ref {args.reference!r} {vcfFile!r} > {consensusFile!r}"
        )

        if not args.dryRun and result.stderr:
            print(result.stderr, end="", file=sys.stderr)

    if not args.dryRun:
        consensus = list(FastaReads(consensusFile))[0]
        if args.id is not None:
            consensus.id = args.id
        elif args.idLambda is not None:
            idLambda = eval(args.idLambda)
            consensus.id = idLambda(consensus.id)

        print(consensus.toString("fasta"), end="")

    if args.dryRun or args.log:
        print("\n".join(e.log), file=sys.stderr)

    if tempdir:
        if args.clean:
            e.execute(f"rm -r {tempdir!r}")
        else:
            print(f"Temporary directory {tempdir!r}.", file=sys.stderr)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import argparse
import os
import sys
import time
from os.path import exists

import numpy as np
import pandas as pd
import plotly.express as px
from sklearn.manifold import MDS

from dark.sam import DistanceMatrix
from dark.utils import readLabels


# The following function is Hepatitis B virus (HBV) specific. So it really
# has no place being in the dark-matter repo. But it's convenient to have
# it here and I don't for the moment have a better (and certainly not
# simpler) place to put it and make it easily accessible (without adding
# another Python dependency), so here it sits, at least for now.
def HBVGenotypeKey(genotype):
    """
    Make a key for sorting HBV genotypes (including non-human genotypes that
    are the species name) for the plot legend.

    @param genotype: A C{str} genotype.
    @return: A 2-C{tuple} with an C{int} index and the passed C{str} genotype.
        This will ensure that sorting will be first on the index, then on the
        genotype name.
    """
    # The index is 0 for single-character genotypes (apart from ?, which
    # comes last), followed by 1 for genotypes with longer names
    # (Chimpanzee, Gibbon, etc), followed by 2 for unknown (?) genotypes.
    return (2 if genotype == "?" else (0 if len(genotype) == 1 else 1), genotype)


def getFigure(referenceIds, categories, dm, transform, args):
    """
    Make a 3D figure.

    @param referenceIds: A C{list} of C{str} reference ids.
    @param categories: A C{list} of C{str} reference id categories.
    @param dm: A C{DistanceMatrix} instance.
    @param transform: An C{MDS} transform.
    @param args: A C{Namespace} object returned by argparse, containing
        argument values passed on the command line.
    @return: A plotly express figure instance.
    """
    readCounts = [len(dm.scores[id_]) for id_ in referenceIds]

    if not (len(transform) == len(readCounts) == len(categories) == len(referenceIds)):
        raise ValueError(
            f"Unequal lengths: transform={len(transform)}, "
            f"readCounts={len(readCounts)}, categories={len(categories)}, "
            f"referenceIds={len(referenceIds)}."
        )

    if args.categorySortKey == "HBV genotype":
        categorySortKey = HBVGenotypeKey
    else:

        def categorySortKey(c):
            return c

    df = pd.DataFrame(
        {
            "x": transform[:, 0],
            "y": transform[:, 1],
            "Read count": readCounts,
            args.categoryName: categories,
            "Accession": referenceIds,
        }
    )

    categoryOrders = {args.categoryName: sorted(categories, key=categorySortKey)}

    if args.twoD:
        fig = px.scatter(
            df,
            x="x",
            y="y",
            color=args.categoryName,
            opacity=0.70,
            title="Reference map",
            hover_data=("Accession", "Read count"),
            category_orders=categoryOrders,
        )
    else:
        fig = px.scatter_3d(
            df,
            x="x",
            y="y",
            z="Read count",
            color=args.categoryName,
            opacity=0.70,
            title="Reference map",
            hover_data=("Accession",),
            category_orders=categoryOrders,
        )

    fig.update_traces(marker={"size": 4.2})

    fig.update_layout(
        margin=dict(l=0, r=0, b=0, t=0),
        legend=dict(yanchor="top", y=0.99, xanchor="left", x=0.01),
    )

    if args.logZ and not args.twoD:
        fig.update_layout(scene_zaxis_type="log")

    return fig


def getTransform(referenceIds, dm, args):
    """
    Get the MDS-transformed locations of the references.

    @param referenceIds: A C{list} of C{str} reference ids.
    @param dm: A C{DistanceMatrix} instance.
    @param args: A C{Namespace} object returned by argparse, containing
        argument values passed on the command line.
    @return: A numpy n x 2 array of x, y locations (where n is the number of
        reference ids).
    """
    if args.mdsFile and exists(args.mdsFile):
        if args.verbose:
            modificationTime = time.strftime(
                "%Y-%m-%d %H:%M:%S", time.localtime(os.path.getmtime(args.mdsFile))
            )
            print(
                f"Loading cached MDS transform file dated "
                f"{modificationTime} from {args.mdsFile!r}.",
                file=sys.stderr,
            )
        transform = np.load(args.mdsFile)
    else:
        distance = dm.matrix(
            referenceIds=referenceIds, metric=args.metric, similarity=False
        )
        mds = MDS(dissimilarity="precomputed", n_jobs=8)
        if args.verbose:
            print("Starting MDS fit.", file=sys.stderr)
        transform = mds.fit_transform(distance)
        if args.verbose:
            print("Finished MDS fit.", file=sys.stderr)

        # Cache the result if an MDS file name was given.
        if args.mdsFile:
            with open(args.mdsFile, "wb") as fp:
                np.save(fp, transform)

    return transform


def getDistanceMatrix(args):
    """
    Get the distance matrix for the references.

    @param args: A C{Namespace} object returned by argparse, containing
        argument values passed on the command line.
    @return: A C{DistanceMatrix} instance.
    """
    dm = DistanceMatrix()

    if args.samFile:
        for filename in args.samFile:
            dm.addFile(filename, scoreTag="AS")

        # Cache the distance matrix scores if a file name was given.
        if args.scoreFile:
            with open(args.scoreFile, "w") as fp:
                dm.save(fp)
    else:
        # The argparse group should ensure that --scoreFile was used if
        # --samFile was not.
        assert args.scoreFile

        if exists(args.scoreFile):
            with open(args.scoreFile) as fp:
                dm.load(fp)
        else:
            print("Score file {args.scoreFile!r} does not exist.", file=sys.stderr)
            sys.exit(1)

    return dm


def main(args):
    """
    Plot references according to their distances from each other, based on
    common read matches (and non-matches).

    @param args: A C{Namespace} object returned by argparse, containing
        argument values passed on the command line.
    """
    if not (args.samFile or args.scoreFile):
        print(
            "At least a SAM file or a score file name must be given.", file=sys.stderr
        )
        sys.exit(1)

    dm = getDistanceMatrix(args)

    if args.minMatchingReads is None:
        referenceIds = sorted(dm.scores)
    else:
        referenceIds = sorted(
            referenceId
            for (referenceId, reads) in dm.scores.items()
            if len(reads) >= args.minMatchingReads
        )
        if args.verbose:
            print(
                f"Found {len(referenceIds)} references with at least "
                f"{args.minMatchingReads} matching reads.",
                file=sys.stderr,
            )

    if args.categoryFile:
        with open(args.categoryFile) as fp:
            categoryDict = readLabels(fp)
        categories = [categoryDict.get(id_, id_) for id_ in referenceIds]
    else:
        categories = referenceIds

    # This plots a dendrogram - not needed / wanted at the moment.
    #
    # from matplotlib import pyplot as plt
    # from scipy.cluster.hierarchy import dendrogram, linkage
    # from scipy.spatial.distance import squareform
    #
    # condensed = squareform(distance)
    # linkage_ = linkage(condensed, method='average')
    # dendrogram(linkage_, orientation='left', labels=categories,
    #            leaf_font_size=2)
    # plt.show()

    transform = getTransform(referenceIds, dm, args)

    fig = getFigure(referenceIds, categories, dm, transform, args)

    fig.write_html(args.htmlFile)
    if args.verbose:
        print(f"Wrote {args.htmlFile!r}.", file=sys.stderr)


def makeParser():
    """
    Make a command-line argument parser.

    @return: An C{argparse.ArgumentParser} instance.
    """
    parser = argparse.ArgumentParser(
        description=(
            "Plot references according to their distances from one "
            "another, based on common read matches (and "
            "non-matches)."
        )
    )

    parser.add_argument(
        "--samFile", action="append", help="The SAM file(s) to load. May be repeated."
    )

    parser.add_argument(
        "--scoreFile",
        help=(
            "A (JSON) file to read (JSON) the reference/read score matrix "
            "from. This should have been produced by an earlier call with a "
            "--scoreFile argument or else by saving the output of "
            "reference-read-scores-to-JSON.py  If the file does not exist, "
            "it will be created."
        ),
    )

    parser.add_argument(
        "--htmlFile", required=True, help="The HTML plotly file to write."
    )

    parser.add_argument(
        "--mdsFile", help="A JSON to cache the result of the MDS optimization."
    )

    parser.add_argument(
        "--categoryName",
        default="Genotype",
        help=(
            "The name of the categories to which reference sequences are "
            'assigned (e.g., "Genotype"). This name will appear at the '
            "top of the legend."
        ),
    )

    parser.add_argument(
        "--categorySortKey",
        choices=("HBV genotype",),
        help=(
            "The name of the function to use to sort categories (for the "
            "legend). If not given, categories will be sorted by name."
        ),
    )

    parser.add_argument(
        "--categoryFile",
        help=(
            "A file containing labels for reference sequences. The format "
            "is lines containing a reference name, a TAB, and a new name "
            "(to be shown in the plot). If no label file is given, or if "
            "there is no new name for a reference, the original reference "
            "names will be used."
        ),
    )

    parser.add_argument(
        "--minMatchingReads",
        type=int,
        help=(
            "The minimum number of reads that must match a reference for it "
            "to be included."
        ),
    )

    parser.add_argument(
        "--verbose", action="store_true", help="Print extra information."
    )

    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        "--2d",
        "--2D",
        action="store_true",
        dest="twoD",
        help=(
            "Make the figure in two dimensions, instead of three. The third "
            "dimensions (reference matching read count) is omitted."
        ),
    )

    group.add_argument(
        "--logZ",
        action="store_true",
        help="Log the z (read count) axis (only valid if --2d is not used).",
    )

    parser.add_argument(
        "--metric",
        choices=("jaccard", "soergel"),
        default="soergel",
        help=("The distance metric to use."),
    )

    return parser


if __name__ == "__main__":
    main(makeParser().parse_args())



#!/usr/bin/env python

import argparse
import sys

from ete4 import Tree


def readNames(fp):
    """
    Read the existing / news names file.

    @param fp: A file pointer containing renaming information.

    @raise ValueError: If there is a duplicate old name given or if a line
        cannot be split into two TAB-separated fields.
    @return: A C{dict} of old to new names.
    """
    result = {}
    for line in fp:
        old, new = line.strip().split("\t")
        if old in result:
            raise ValueError(f"Sequence id {old!r} appears twice in the input.")
        else:
            result[old] = new

    return result


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description="Relabel a Newick tree according to a file of from/to names.",
)

parser.add_argument("newickFile", help="The Newick tree file.")

parser.add_argument(
    "--nameFile",
    help=(
        "The file of renamings. Each line should contain an existing (in "
        "the tree) sequence id, a TAB, and then a new sequence id. If not "
        "given, the renaming is read from standard input."
    ),
)

parser.add_argument(
    "--inputFormat",
    type=int,
    default=0,
    help=(
        "The parser argument to pass to the ete Tree class initializer to "
        "read the input Newick."
    ),
)

parser.add_argument(
    "--outputFormat",
    type=int,
    default=0,
    help="The parser argument to pass to the ete Tree writer for the resulting tree.",
)

parser.add_argument(
    "--quotedNames",
    action="store_true",
    help="If given, pass quoted_node_names=True to the ete Tree class initializer",
)

parser.add_argument(
    "--verbose",
    action="store_true",
    help="Print details of the renaming(s).",
)

args = parser.parse_args()

tree = Tree(
    args.newickFile, parser=args.inputFormat, quoted_node_names=args.quotedNames
)

if args.nameFile:
    with open(args.nameFile) as fp:
        names = readNames(fp)
else:
    names = readNames(sys.stdin)


# Relabel tree nodes that are mentioned in args.nameFile.
for node in tree.traverse():
    try:
        newName = names[node.name]
    except KeyError:
        pass
    else:
        if args.verbose:
            print(f"Renaming {node.name!r} to {newName!r}", file=sys.stderr)
        node.name = newName

# Print the tree, including the name of the root.
print("%s%s;" % (tree.write(parser=args.outputFormat)[:-1], tree.name))



from unittest import TestCase

from dark.html import NCBISequenceLink, NCBISequenceLinkURL


class TestNCBISequenceLinkURL(TestCase):
    """
    Test the NCBISequenceLinkURL function.
    """

    def testNoField(self):
        """
        If no field is passed, the passed title must be used.
        """
        self.assertEqual(
            "http://www.ncbi.nlm.nih.gov/nuccore/xxx", NCBISequenceLinkURL("xxx")
        )

    def testFieldNumber(self):
        title = "gi|323924|gb|M15204.1|FCVMYCCA Feline leukemia virus myc gene"
        self.assertEqual(
            "http://www.ncbi.nlm.nih.gov/nuccore/M15204.1",
            NCBISequenceLinkURL(title, 3),
        )

    def testAlternateDelimiter(self):
        title = "gi+37955203+gb+AY253278.1+ Homo sapiens clone AL-11 HIV-1"
        self.assertEqual(
            "http://www.ncbi.nlm.nih.gov/nuccore/AY253278.1",
            NCBISequenceLinkURL(title, 3, "+"),
        )

    def testFieldOutOfRange(self):
        """
        If the field number passed does not exist, IndexError must be raised.
        """
        title = "gi|37955203|gb|AY253278.1| Homo sapiens clone AL-11 HIV-1"
        error = (
            r"^Could not extract field 10 from sequence title "
            r"'gi\|37955203\|gb\|AY253278\.1\| Homo sapiens clone "
            "AL-11 HIV-1'$"
        )
        self.assertRaisesRegex(IndexError, error, NCBISequenceLinkURL, title, 10)


class TestNCBISequenceLink(TestCase):
    """
    Test the NCBISequenceLink function.
    """

    def testNoField(self):
        """
        If no field is passed, the passed title must be used.
        """
        self.assertEqual(
            '<a href="http://www.ncbi.nlm.nih.gov/nuccore/xxx" '
            + 'target="_blank">xxx</a>',
            NCBISequenceLink("xxx"),
        )

    def testFieldNumber(self):
        title = "gi|323924|gb|M15204.1|FCVMYCCA Feline leukemia virus myc gene"
        self.assertEqual(
            '<a href="http://www.ncbi.nlm.nih.gov/nuccore/M15204.1" '
            + 'target="_blank">'
            + title
            + "</a>",
            NCBISequenceLink(title, 3),
        )

    def testAlternateDelimiter(self):
        title = "gi+37955203+gb+AY253278.1+ Homo sapiens clone AL-11 HIV-1"
        self.assertEqual(
            '<a href="http://www.ncbi.nlm.nih.gov/nuccore/AY253278.1" '
            + 'target="_blank">'
            + title
            + "</a>",
            NCBISequenceLink(title, 3, "+"),
        )

    def testFieldOutOfRange(self):
        """
        If the field number passed does not exist, IndexError must be raised.
        """
        title = "gi|37955203|gb|AY253278.1| Homo sapiens clone AL-11 HIV-1"
        error = (
            r"^Could not extract field 10 from sequence title "
            r"'gi\|37955203\|gb\|AY253278\.1\| Homo sapiens clone "
            "AL-11 HIV-1'$"
        )
        self.assertRaisesRegex(IndexError, error, NCBISequenceLink, title, 10)



#!/usr/bin/env python

import sys

from dark import fasta

if __name__ == "__main__":
    if len(sys.argv) == 1:
        print(
            "Usage: %s 1.fasta, 2.fasta, ... > seq.fasta" % (sys.argv[0]),
            file=sys.stderr,
        )
        sys.exit(1)
    else:
        for read in fasta.fastaSubtract(sys.argv[1:]):
            print(read.toString(), end="")
